"""Functions managing the dict of formulas used by Biogeme


:author: Michel Bierlaire

:date: Mon May 13 13:42:50 2024

"""

import logging
from typing import Any

from fuzzywuzzy import fuzz

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression

logger = logging.getLogger(__name__)

# Valid names for the log likelihood.  Only one of them can be present in the dict.
log_like_names = ['log_like', 'loglike']

# Valid names for the weights.  Only one of them can be present in the dict.
weight_names = ['log_like', 'loglike']


def check_validity(dict_of_formulas: dict[str, Any]) -> None:
    """Verifies if the formulas are Biogeme expressions. If not, an exception is raised"""
    for k, f in dict_of_formulas.items():
        if not isinstance(f, Expression):
            raise BiogemeError(
                f'Expression for "{k}" is not of type '
                f"biogeme.expressions.Expression. "
                f"It is of type {type(f)}"
            )


def is_similar_to(word_1: str, word_2: str) -> bool:
    """Checks if two words are similar."""
    return fuzz.ratio(word_1, word_2) >= 80


def get_expression(
    dict_of_formulas: dict[str, Expression], valid_keywords: list[str]
) -> Expression | None:
    """Extract the formula for specific keywords

    :param dict_of_formulas: as the name says...
    :param valid_keywords: keywords that are considered valid to represent the expression
    :return: the requested expression
    """
    found_name = None
    for valid_name in valid_keywords:
        the_expression = dict_of_formulas.get(valid_name)
        if the_expression is not None:
            if found_name is not None:
                error_msg = (
                    f"This expression can be defined with the keyword "
                    f"'{found_name}' or '{valid_name}' but not both."
                )
                raise BiogemeError(error_msg)
            found_name = valid_name

    if found_name is None:
        for valid_name in valid_keywords:
            for key in dict_of_formulas:
                if is_similar_to(valid_name, key):
                    warning_msg = f'In the formulas, one key is "{key}". Should it be "{valid_name}" instead?'
                    logger.warning(warning_msg)
        return None
    return dict_of_formulas[found_name]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from biogeme.database import Database
from biogeme.model_elements import ModelElements
from biogeme.second_derivatives import SecondDerivativesMode
from .function_call import (
    CallableExpression,
    CompiledFormulaEvaluator,
    function_from_compiled_formula,
)
from .single_formula import evaluate_formula, evaluate_model_per_row

if TYPE_CHECKING:
    from biogeme.expressions import Expression


def evaluate_simple_expression_per_row(
    expression: Expression,
    database: Database,
    numerically_safe: bool,
    second_derivatives_mode: SecondDerivativesMode,
) -> np.ndarray:
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression, weight=None, database=database
    )
    return evaluate_model_per_row(
        model_elements=model_elements,
        the_betas={},
        numerically_safe=numerically_safe,
        second_derivatives_mode=second_derivatives_mode,
    )


def evaluate_simple_expression(
    expression: Expression, database: Database | None, numerically_safe: bool
) -> float:
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression, weight=None, database=database
    )

    return evaluate_formula(
        model_elements=model_elements,
        the_betas={},
        second_derivatives_mode=SecondDerivativesMode.NEVER,
        numerically_safe=numerically_safe,
    )


def create_function_simple_expression(
    expression: Expression,
    numerically_safe: bool,
    named_output: bool = False,
    database: Database | None = None,
) -> CallableExpression:
    if database is None:
        database = Database.dummy_database()
    model_elements = ModelElements.from_expression_and_weight(
        log_like=expression, weight=None, database=database
    )

    the_evaluator = CompiledFormulaEvaluator(
        model_elements=model_elements,
        second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
        numerically_safe=numerically_safe,
    )
    betas = model_elements.expressions_registry.free_betas_init_values
    return function_from_compiled_formula(
        the_compiled_function=the_evaluator, the_betas=betas, named_output=named_output
    )



"""Test the catalog object

Michel Bierlaire
Thu Apr 17 2025, 08:31:00
"""

import unittest

from biogeme.catalog import Catalog, Controller
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, NamedExpression


class DummyExpression(Expression):
    """Minimal Expression subclass for testing purposes."""

    def __init__(self, label):
        self.label = label

    def get_all_controllers(self):
        return set()


class TestCatalog(unittest.TestCase):
    def test_valid_catalog_creation(self):
        expr1 = DummyExpression("e1")
        expr2 = DummyExpression("e2")
        named_exprs = [
            NamedExpression("A", expr1),
            NamedExpression("B", expr2),
        ]
        catalog = Catalog("myCatalog", named_exprs)
        self.assertIsNotNone(catalog.controlled_by)
        self.assertEqual(catalog.selected().name, "A")
        self.assertIs(catalog.selected().expression, expr1)

    def test_catalog_with_empty_list_raises(self):
        with self.assertRaisesRegex(
            BiogemeError, "cannot create a catalog from an empty list"
        ):
            Catalog("emptyCatalog", [])

    def test_catalog_with_wrong_controller_type_raises(self):
        expr = DummyExpression("e")
        named_exprs = [NamedExpression("A", expr)]
        with self.assertRaisesRegex(
            BiogemeError, "The controller must be of type Controller"
        ):
            Catalog("badController", named_exprs, controlled_by="not a controller")

    def test_controller_name_mismatch_raises(self):
        expr = DummyExpression("e")
        named_exprs = [NamedExpression("A", expr), NamedExpression("B", expr)]
        wrong_controller = Controller("myController", ["X", "Y"])
        with self.assertRaisesRegex(BiogemeError, "Incompatible IDs between catalog"):
            Catalog("mismatch", named_exprs, controlled_by=wrong_controller)

    def test_catalog_from_dict(self):
        expr1 = DummyExpression("e1")
        expr2 = DummyExpression("e2")
        expr_dict = {
            "A": expr1,
            "B": expr2,
        }
        catalog = Catalog.from_dict("dictCatalog", expr_dict)
        self.assertEqual(len(list(catalog.get_iterator())), 2)
        self.assertEqual(catalog.selected().name, "A")
        self.assertIsInstance(catalog.selected().expression, DummyExpression)

    def test_get_all_controllers_returns_controller(self):
        expr = DummyExpression("e")
        named_exprs = [NamedExpression("A", expr)]
        catalog = Catalog("controllerTest", named_exprs)
        controllers = catalog.get_all_controllers()
        self.assertIsInstance(controllers, set)
        self.assertIn(catalog.controlled_by, controllers)



"""
Test the tools module

:author: Michel Bierlaire
:date: Sun Aug  8 17:42:48 2021
"""

import os
import tempfile
import unittest
from copy import deepcopy
from datetime import timedelta
from unittest.mock import patch

import numpy as np
import pandas as pd

import biogeme.tools.database
import biogeme.tools.derivatives
import biogeme.tools.files
import biogeme.tools.likelihood_ratio
import biogeme.tools.primes
import biogeme.tools.unique_ids
from biogeme.exceptions import BiogemeError
from biogeme.function_output import FunctionOutput
from biogeme.tools import (
    create_backup,
    format_timedelta,
    is_valid_filename,
    safe_deserialize_array,
    safe_serialize_array,
)
from biogeme.tools.derivatives import CheckDerivativesResults
from test_data import (
    input_flatten,
    output_flatten_1,
    output_flatten_2,
    output_flatten_3,
)


# Bug in pylint
# pylint: disable=no-member
#
# Too constraining
# pylint: disable=invalid-name, too-many-instance-attributes
#
# Not needed in test
# pylint: disable=missing-function-docstring, missing-class-docstring


def my_function(
    x: np.ndarray, gradient: bool, hessian: bool, bhhh: bool
) -> FunctionOutput:
    f = np.log(x[0]) + np.exp(x[1])
    g = np.empty(2)
    g[0] = 1.0 / x[0]
    g[1] = np.exp(x[1])
    H = np.empty((2, 2))
    H[0, 0] = -1.0 / x[0] ** 2
    H[0, 1] = 0.0
    H[1, 0] = 0.0
    H[1, 1] = np.exp(x[1])
    return FunctionOutput(function=f, gradient=g, hessian=H)


class TestTools(unittest.TestCase):
    def test_findiff_g(self):
        x = np.array([1.1, 1.1])
        g_fd = biogeme.tools.derivatives.findiff_g(my_function, x)
        np.testing.assert_almost_equal(g_fd, [0.90909087, 3.00416619], decimal=5)

    def test_findiff_H(self):
        x = np.array([1.1, 1.1])
        H_fd = biogeme.tools.derivatives.findiff_h(my_function, x)
        np.testing.assert_almost_equal(
            H_fd,
            [
                [
                    -0.8264462,
                    0,
                ],
                [
                    0,
                    3.00416619,
                ],
            ],
            decimal=5,
        )

    def test_checkDerivatives(self):
        x = np.array([1.1, -1.5])
        results: CheckDerivativesResults = biogeme.tools.derivatives.check_derivatives(
            my_function, x, names=['First', 'Second']
        )

        np.testing.assert_almost_equal(results.errors_gradient, [0, 0])
        np.testing.assert_almost_equal(results.errors_hessian, [[0, 0], [0, 0]])

        x = np.array([3.2, 1.32])
        results: CheckDerivativesResults = biogeme.tools.derivatives.check_derivatives(
            my_function, x, names=['First', 'Second']
        )
        np.testing.assert_almost_equal(results.errors_gradient, [0, 0], decimal=5)
        np.testing.assert_almost_equal(
            results.errors_hessian, [[0, 0], [0, 0]], decimal=5
        )

    def test_getPrimeNumbers(self):
        result = biogeme.tools.primes.get_prime_numbers(7)
        self.assertListEqual(result, [2, 3, 5, 7, 11, 13, 17])

        with self.assertRaises(BiogemeError):
            result = biogeme.tools.primes.get_prime_numbers(0)

        with self.assertRaises(BiogemeError):
            result = biogeme.tools.primes.get_prime_numbers(-1)

        with self.assertRaises(BiogemeError):
            result = biogeme.tools.primes.get_prime_numbers(0.3)

    def test_calculatePrimeNumbers(self):
        result = biogeme.tools.primes.calculate_prime_numbers(10)
        self.assertListEqual(result, [2, 3, 5, 7])
        result = biogeme.tools.primes.calculate_prime_numbers(0)
        self.assertListEqual(result, [])

        with self.assertRaises(BiogemeError):
            result = biogeme.tools.primes.calculate_prime_numbers(-1)

        with self.assertRaises(BiogemeError):
            result = biogeme.tools.primes.calculate_prime_numbers(0.3)

    def test_countNumberOfGroups(self):
        df = pd.DataFrame(
            {
                'ID': [1, 1, 2, 3, 3, 1, 2, 3],
                'value': [1000, 2000, 3000, 4000, 5000, 5000, 10000, 20000],
            }
        )
        nbr = biogeme.tools.count_number_of_groups(df, 'ID')
        self.assertEqual(nbr, 6)
        nbr = biogeme.tools.count_number_of_groups(df, 'value')
        self.assertEqual(nbr, 7)

    def test_likelihood_ratio_test(self):
        model1 = (-1340.8, 5)
        model2 = (-1338.49, 7)
        r = biogeme.tools.likelihood_ratio.likelihood_ratio_test(model1, model2)
        expected = (
            'H0 cannot be rejected at level 5.0%',
            4.619999999999891,
            5.991464547107979,
        )
        self.assertEqual(len(r), len(expected))
        self.assertEqual(r[0], expected[0])
        self.assertAlmostEqual(r[1], expected[1], places=3)
        self.assertAlmostEqual(r[2], expected[2], places=3)

        r = biogeme.tools.likelihood_ratio.likelihood_ratio_test(model2, model1)
        self.assertEqual(len(r), len(expected))
        self.assertEqual(r[0], expected[0])
        self.assertAlmostEqual(r[1], expected[1], places=3)
        self.assertAlmostEqual(r[2], expected[2], places=3)
        r = biogeme.tools.likelihood_ratio.likelihood_ratio_test(
            model1, model2, significance_level=0.1
        )
        expected = (
            'H0 can be rejected at level 10.0%',
            4.619999999999891,
            4.605170185988092,
        )

        self.assertEqual(len(r), len(expected))
        self.assertEqual(r[0], expected[0])
        self.assertAlmostEqual(r[1], expected[1], places=3)
        self.assertAlmostEqual(r[2], expected[2], places=3)
        with self.assertRaises(BiogemeError):
            model1 = (-1340.8, 7)
            model2 = (-1338.49, 5)
            biogeme.tools.likelihood_ratio.likelihood_ratio_test(model1, model2)

    def test_flatten_database(self):
        df = deepcopy(input_flatten)
        result_1 = biogeme.tools.database.flatten_database(df, 'ID', row_name='Name')
        output_flatten_1.index.name = 'ID'
        pd.testing.assert_frame_equal(result_1, output_flatten_1)
        result_2 = biogeme.tools.database.flatten_database(input_flatten, 'ID')
        output_flatten_2.index.name = 'ID'
        pd.testing.assert_frame_equal(result_2, output_flatten_2)
        result_3 = biogeme.tools.database.flatten_database(
            df, 'ID', identical_columns=[]
        )
        output_flatten_3.index.name = 'ID'
        pd.testing.assert_frame_equal(result_3, output_flatten_3)

    def test_temporary_file(self):
        content = 'a random sentence'
        with biogeme.tools.files.TemporaryFile() as temporary_file:
            with open(temporary_file.name, 'w', encoding='utf-8') as f:
                f.write(content)
            with open(temporary_file.name, encoding='utf-8') as f:
                check = f.read()
            self.assertEqual(content, check)

    def test_model_names(self):
        model_names = biogeme.tools.unique_ids.ModelNames(prefix='Test')

        name_a = model_names('model_a')
        correct_name_a = 'Test_000000'
        self.assertEqual(name_a, correct_name_a)
        name_b = model_names('model_b')
        correct_name_b = 'Test_000001'
        self.assertEqual(name_a, correct_name_a)
        name_a = model_names('model_a')
        self.assertEqual(name_a, correct_name_a)

    def test_generate_unique_ids(self):
        list_1 = ['A', 'B', 'C']
        correct_dict_1 = {'A': 'A', 'B': 'B', 'C': 'C'}
        dict_1 = biogeme.tools.unique_ids.generate_unique_ids(list_1)
        self.assertDictEqual(correct_dict_1, dict_1)

        list_2 = ['A', 'B', 'B']
        correct_dict_2 = {'A': 'A', 'B_0': 'B', 'B_1': 'B'}
        dict_2 = biogeme.tools.unique_ids.generate_unique_ids(list_2)
        self.assertDictEqual(correct_dict_2, dict_2)

        list_3 = ['B', 'B', 'B']
        correct_dict_3 = {'B_2': 'B', 'B_0': 'B', 'B_1': 'B'}
        dict_3 = biogeme.tools.unique_ids.generate_unique_ids(list_3)
        self.assertDictEqual(correct_dict_3, dict_3)


class UniqueProductTestCase(unittest.TestCase):
    def test_unique_product(self):
        # Test case with small iterables
        iterables = [range(2), range(3), range(2)]
        expected_output = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (0, 2, 0),
            (0, 2, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
            (1, 2, 0),
            (1, 2, 1),
        ]
        result = list(biogeme.tools.unique_ids.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with larger iterables
        iterables = [range(3), range(4), range(2)]
        expected_output = [
            (0, 0, 0),
            (0, 0, 1),
            (0, 1, 0),
            (0, 1, 1),
            (0, 2, 0),
            (0, 2, 1),
            (0, 3, 0),
            (0, 3, 1),
            (1, 0, 0),
            (1, 0, 1),
            (1, 1, 0),
            (1, 1, 1),
            (1, 2, 0),
            (1, 2, 1),
            (1, 3, 0),
            (1, 3, 1),
            (2, 0, 0),
            (2, 0, 1),
            (2, 1, 0),
            (2, 1, 1),
            (2, 2, 0),
            (2, 2, 1),
            (2, 3, 0),
            (2, 3, 1),
        ]
        result = list(biogeme.tools.unique_ids.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with empty iterable
        iterables = [range(3), [], range(2)]
        expected_output = []
        result = list(biogeme.tools.unique_ids.unique_product(*iterables))
        self.assertEqual(result, expected_output)

        # Test case with single iterable
        iterables = [range(3)]
        expected_output = [(0,), (1,), (2,)]
        result = list(biogeme.tools.unique_ids.unique_product(*iterables))
        self.assertEqual(result, expected_output)


class TestFormatTimedelta(unittest.TestCase):

    def test_hours_minutes_seconds(self):
        td = timedelta(hours=2, minutes=30, seconds=15)
        self.assertEqual(format_timedelta(td), '2h 30m 15s')

    def test_minutes_seconds(self):
        td = timedelta(minutes=45, seconds=30)
        self.assertEqual(format_timedelta(td), '45m 30s')

    def test_seconds_microseconds(self):
        td = timedelta(seconds=5, microseconds=500000)
        self.assertEqual(format_timedelta(td), '5.5s')

    def test_only_microseconds_more_than_a_millisecond(self):
        td = timedelta(microseconds=1250)
        self.assertEqual(format_timedelta(td), '1ms')

    def test_only_microseconds_less_than_a_millisecond(self):
        td = timedelta(microseconds=999)
        self.assertEqual(format_timedelta(td), '999μs')

    def test_zero_units(self):
        td = timedelta(hours=0, minutes=0, seconds=5)
        self.assertEqual(format_timedelta(td), '5.0s')

    def test_large_timedelta(self):
        td = timedelta(days=2, hours=3, minutes=4, seconds=5)
        self.assertEqual(format_timedelta(td), '51h 4m 5s')

    def test_microseconds_rounding(self):
        # This test ensures that microseconds are correctly floored when converted to milliseconds
        td = timedelta(microseconds=1999)  # Should floor to 1ms, not round to 2ms
        self.assertEqual(format_timedelta(td), '1ms')


class TestIsValidFilename(unittest.TestCase):

    def test_empty_filename(self):
        self.assertEqual(is_valid_filename(''), (False, 'Name is empty'))

    def test_filename_with_invalid_chars(self):
        invalid_filenames = ['<invalid>', 'in"valid', 'in|valid', 'invalid?']
        for filename in invalid_filenames:
            result, message = is_valid_filename(filename)
            self.assertFalse(result)
            self.assertIn('Name contains one invalid char:', message)

    def test_windows_reserved_names(self):
        if os.name == 'nt':
            reserved_names = ["CON", "PRN", "AUX", "NUL", "COM1", "LPT1"]
            for name in reserved_names:
                result, message = is_valid_filename(name)
                self.assertFalse(result)
                self.assertIn('Name is a reserved name:', message)

    def test_exceeds_length_limit(self):
        long_filename = 'a' * 256
        self.assertEqual(
            is_valid_filename(long_filename),
            (False, f'The length of the filename exceeds 255: {len(long_filename)}'),
        )

    def test_valid_filename(self):
        self.assertEqual(is_valid_filename('valid_filename.txt'), (True, ''))

    def test_filename_with_space(self):
        self.assertEqual(is_valid_filename('valid filename.txt'), (True, ''))


class TestCreateBackup(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for the tests
        self.test_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.test_dir.cleanup)  # Ensure cleanup after tests

    def create_test_file(self, filename, content='test content'):
        """Helper function to create a file with specified content."""
        path = os.path.join(self.test_dir.name, filename)
        with open(path, 'w') as f:
            f.write(content)
        return path

    @patch('biogeme.tools.files.logger')
    def test_file_does_not_exist(self, mock_logger):
        non_existent_file = os.path.join(self.test_dir.name, 'non_existent.txt')
        self.assertFalse(os.path.exists(non_existent_file))
        result = create_backup(non_existent_file)
        # Get the arguments of the last call to mock_logger.info
        args, kwargs = mock_logger.info.call_args

        # Check if the first argument (the log message) contains the substring
        self.assertIn(
            'No backup has been generated',
            args[0],
            "Log message does not contain expected substring",
        )

    def test_file_exists_rename_true(self):
        filename = self.create_test_file('exists.txt')
        self.assertTrue(os.path.exists(filename))
        backup_name = create_backup(filename)
        self.assertFalse(os.path.exists(filename))
        self.assertTrue(os.path.exists(backup_name))

    def test_file_exists_rename_false(self):
        filename = self.create_test_file('exists_copy.txt')
        self.assertTrue(os.path.exists(filename))
        backup_name = create_backup(filename, rename=False)
        self.assertTrue(os.path.exists(filename))
        self.assertTrue(os.path.exists(backup_name))

    def test_multiple_backups(self):
        filename = self.create_test_file('multiple_backups.txt')
        backup_name1 = create_backup(filename, rename=False)
        backup_name2 = create_backup(filename, rename=False)
        self.assertNotEqual(backup_name1, backup_name2)
        self.assertTrue(os.path.exists(filename))
        self.assertTrue(os.path.exists(backup_name1))
        self.assertTrue(os.path.exists(backup_name2))


class TestSafeSerialization(unittest.TestCase):
    def test_serialize_deserialize_1d(self):
        arr = np.array([1.0, np.nan, 3.0])
        serialized = safe_serialize_array(arr)
        self.assertEqual(serialized, [1.0, None, 3.0])
        deserialized = safe_deserialize_array(serialized)
        np.testing.assert_array_equal(np.isnan(arr), np.isnan(deserialized))
        np.testing.assert_allclose(
            np.nan_to_num(arr, nan=0.0),
            np.nan_to_num(deserialized, nan=0.0),
            rtol=1e-10,
        )

    def test_serialize_deserialize_2d(self):
        arr = np.array([[1.0, np.nan], [2.0, 3.0]])
        serialized = safe_serialize_array(arr)
        self.assertEqual(serialized, [[1.0, None], [2.0, 3.0]])
        deserialized = safe_deserialize_array(serialized)
        np.testing.assert_array_equal(np.isnan(arr), np.isnan(deserialized))
        np.testing.assert_allclose(
            np.nan_to_num(arr, nan=0.0),
            np.nan_to_num(deserialized, nan=0.0),
            rtol=1e-10,
        )

    def test_invalid_input_serialize(self):
        with self.assertRaises(TypeError):
            safe_serialize_array("not an array")

    def test_invalid_input_deserialize(self):
        with self.assertRaises(TypeError):
            safe_deserialize_array("not a list")


if __name__ == '__main__':
    unittest.main()



"""Audit procedure for arithmetic expressions

Michel Bierlaire
Fri Mar 28 17:01:01 2025
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from biogeme.audit_tuple import AuditTuple

from .belongs_to import BelongsTo
from .comparison_expressions import ComparisonOperator
from .draws import Draws
from .integrate import IntegrateNormal
from .logit_expressions import LogLogit
from .montecarlo import MonteCarlo
from .random_variable import RandomVariable
from .visitor import ExpressionVisitor

_audit_visitor = ExpressionVisitor()
register_audit = _audit_visitor.register

if TYPE_CHECKING:
    from .base_expressions import Expression


def audit_expression(expr: Expression) -> AuditTuple:
    """
    Audits an expression tree for structural consistency.

    :param expr: The root expression to audit.
    :return: A tuple containing two lists:
             - error_messages: list of error strings.
             - warning_messages: list of warning strings.
    """
    error_messages = []
    warning_messages = []
    ancestors = []
    context = {
        "errors": error_messages,
        "warnings": warning_messages,
        "ancestors": ancestors,  # Stack of ancestor expressions
    }
    _audit_visitor.visit(expr, context)
    return AuditTuple(errors=error_messages, warnings=warning_messages)


def audit_default(expr: Expression, context: dict[str, list[str]]) -> None:
    """
    Default audit function for expressions that do not have a specific audit function.

    :param expr: The current expression node being audited.
    :param context: Dictionary to collect error and warning messages.
    """
    if hasattr(expr, 'left') and hasattr(expr, 'right'):
        if isinstance(expr, ComparisonOperator):
            if isinstance(expr.left, ComparisonOperator) or isinstance(
                expr.right, ComparisonOperator
            ):
                warning = (
                    f"The expression [{expr}] may be a chained comparison (e.g., a <= b <= c), "
                    f"which Biogeme interprets as nested comparisons, not as Python-style chains. "
                    f"Consider splitting it: (a <= b) & (b <= c)."
                )
                context["warnings"].append(warning)


@register_audit(MonteCarlo)
def audit_montecarlo(expr: MonteCarlo, context: dict[str, list[str]]) -> None:
    """
    Audits a MonteCarlo expression for structural consistency.

    :param expr: The MonteCarlo expression to audit.
    :param context: Dictionary to collect error and warning messages.
    """
    if not expr.embed_expression(Draws):
        context["errors"].append(
            f'MonteCarlo expression {repr(expr)} does not contain any Draws expression.'
        )
    if any(child.embed_expression(MonteCarlo) for child in expr.get_children()):
        context["errors"].append(
            f'MonteCarlo expression {repr(expr)} cannot contain another MonteCarlo expression.'
        )


@register_audit(IntegrateNormal)
def audit_integrate(expr: IntegrateNormal, context: dict[str, list[str]]) -> None:
    """
    Audits an Integrate expression for structural consistency.

    :param expr: The Integrate expression to audit.
    :param context: Dictionary to collect error and warning messages.
    """
    if not expr.embed_expression(RandomVariable):
        context["warnings"].append(
            f'Integrate expression {repr(expr)} does not contain any RandomVariable expression.'
        )


@register_audit(BelongsTo)
def audit_belongsto(expr: BelongsTo, context: dict[str, list[str]]) -> None:
    """
    Audits a BelongsTo expression for structural consistency.

    :param expr: The BelongsTo expression to audit.
    :param context: Dictionary to collect error and warning messages.
    """
    if not all(float(x).is_integer() for x in expr.the_set):
        the_warning = (
            f'The set of numbers used in the expression "BelongsTo" contains '
            f'numbers that are not integer. If it is the intended use, ignore '
            f'this warning: {expr.the_set}.'
        )
        context["warnings"].append(the_warning)


@register_audit(LogLogit)
def audit_loglogit(expr: LogLogit, context: dict[str, list[str]]) -> None:
    """
    Audits a LogLogit expression for structural consistency.

    :param expr: The LogLogit expression to audit.
    :param context: Dictionary to collect error and warning messages.
    """
    if expr.av is None:
        return
    if expr.util.keys() != expr.av.keys():
        the_error = 'Incompatible list of alternatives in logit expression. '
        my_set = expr.util.keys() - expr.av.keys()
        if my_set:
            my_set_content = ', '.join(f'{str(k)} ' for k in my_set)
            the_error += (
                'Id(s) used for utilities and not for availabilities: '
            ) + my_set_content
        my_set = expr.av.keys() - expr.util.keys()
        if my_set:
            my_set_content = ', '.join(f'{str(k)} ' for k in my_set)
            the_error += (
                ' Id(s) used for availabilities and not for utilities: '
            ) + my_set_content
        context["errors"].append(the_error)


@register_audit(RandomVariable)
def audit_randomvariable(expr: RandomVariable, context: dict[str, list[str]]) -> None:
    if not any(
        isinstance(ancestor, IntegrateNormal) for ancestor in context['ancestors']
    ):
        context['errors'].append(
            f'RandomVariable {repr(expr)} is not embedded inside an IntegrateNormal expression.'
        )


@register_audit(Draws)
def audit_draws(expr: Draws, context: dict[str, list[str]]) -> None:
    if not any(isinstance(ancestor, MonteCarlo) for ancestor in context['ancestors']):
        context['errors'].append(
            f'Draws {repr(expr)} is not embedded inside a MonteCarlo expression.'
        )



"""

biogeme.cnl
===========

Example of usage of the cnl module.  This is for programmers who need
examples of use of the functions of the class. The examples are
designed to illustrate the syntax.

:author: Michel Bierlaire
:date: Fri Nov 17 08:27:24 2023

"""

import numpy as np
import pandas as pd
import biogeme.cnl as cnl
import biogeme.tools as tools
import biogeme.biogeme_logging as blog
from biogeme.nests import OneNestForCrossNestedLogit, NestsForCrossNestedLogit

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Logging on')

# %%
# Definition of the nests.
choice_set = ['i', 'j', 'k', 'ell']
mu_nest_1 = 1.4
alphas_1 = {'i': 1, 'j': 0.5, 'k': 0.2}
nest_1 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_1, dict_of_alpha=alphas_1, name='Nest 1'
)
mu_nest_2 = 1.2
alphas_2 = {'j': 0.5, 'k': 0.8, 'ell': 1}
nest_2 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_2, dict_of_alpha=alphas_2, name='Nest 2'
)
nests = NestsForCrossNestedLogit(choice_set=choice_set, tuple_of_nests=(nest_1, nest_2))

# %%
# We retrieve the G function of the cross-nested logit, and verify
# numerically the implementation of the derivatives.

# %%
G = cnl.cnl_G(choice_set, nests)

# %%
# Draw a random point where to evaluate the function.
y = np.random.uniform(low=0.01, high=2, size=4)
y

# %%
f, g, h, gdiff, hdiff = tools.checkDerivatives(G, y, names=None, logg=True)
f

# %%
pd.DataFrame(g)

# %%
pd.DataFrame(h)

# %%
pd.DataFrame(gdiff)

# %%
pd.DataFrame(hdiff)


# %%
# We do the same for the CDF.

# %%
xi = np.random.uniform(low=-10, high=10, size=4)
xi

# %%
F = cnl.cnl_CDF(choice_set, nests)

# %%
f, g, h, gdiff, hdiff = tools.checkDerivatives(F, y, names=None, logg=True)
f

# %%
pd.DataFrame(g)

# %%
pd.DataFrame(h)

# %%
pd.DataFrame(gdiff)

# %%
pd.DataFrame(hdiff)



import os
import unittest
from datetime import timedelta

from biogeme.results_processing import (
    EstimateVarianceCovariance,
    EstimationResults,
    RawEstimationResults,
    format_real_number,
    generate_html_file,
    get_html_condition_number,
    get_html_correlation_results,
    get_html_estimated_parameters,
    get_html_footer,
    get_html_general_statistics,
    get_html_header,
    get_html_preamble,
)
from biogeme.version import get_version, versionDate


class TestHTMLGeneration(unittest.TestCase):

    def setUp(self):
        # Setting up real estimation results using RawEstimationResults
        raw_estimation = RawEstimationResults(
            model_name='test_model',
            user_notes='user notes',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.2, 1.3],
            lower_bounds=[-10, -11],
            upper_bounds=[None, 23],
            gradient=[0.001, -0.002],
            hessian=[[1, 2], [3, 4]],
            bhhh=[[10, 20], [30, 40]],
            null_log_likelihood=-1000,
            initial_log_likelihood=-2000,
            final_log_likelihood=-500,
            data_name='test_data',
            sample_size=1234,
            number_of_observations=987,
            monte_carlo=True,
            number_of_draws=10_000,
            types_of_draws={'var1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=10),
            optimization_messages={'Diagnostic': 'test'},
            convergence=False,
            bootstrap=[[1.0, 1.0], [2.0, 2.0]],
            bootstrap_time=timedelta(seconds=20),
        )

        self.estimation_results = EstimationResults(
            raw_estimation_results=raw_estimation
        )

    def test_format_real_number(self):
        self.assertEqual(format_real_number(123456.789), '1.23e+05')
        self.assertEqual(format_real_number(0.000123), '0.000123')

    def test_get_html_header(self):
        result = get_html_header(self.estimation_results)
        self.assertIn('<html>', result)
        self.assertIn('<title>test_model', result)
        self.assertIn(f'Report from biogeme {get_version()} [{versionDate}]', result)

    def test_get_html_footer(self):
        result = get_html_footer()
        self.assertEqual(result, '</body>\n</html>')

    def test_get_html_preamble(self):
        result = get_html_preamble(self.estimation_results, "testfile.html")
        self.assertIn("testfile.html", result)
        self.assertIn("test_data", result)
        self.assertIn("Algorithm failed to converge", result)

    def test_get_html_general_statistics(self):
        result = get_html_general_statistics(self.estimation_results)
        self.assertIn('<strong>Final log likelihood</strong>', result)
        self.assertIn('<strong>Number of draws</strong>', result)
        self.assertIn('<strong>Sample size</strong>', result)

    def test_get_html_estimated_parameters(self):
        result = get_html_estimated_parameters(
            self.estimation_results,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )
        self.assertIn('<th>Name</th>', result)
        self.assertIn('<td>beta1</td>', result)
        self.assertIn('<td>1.2</td>', result)

    def test_get_html_condition_number(self):
        result = get_html_condition_number(self.estimation_results)
        self.assertIn('<p>Smallest eigenvalue', result)
        self.assertIn('<p>Largest eigenvalue', result)

    def test_get_html_correlation_results(self):
        result = get_html_correlation_results(
            self.estimation_results,
            variance_covariance_type=EstimateVarianceCovariance.ROBUST,
        )
        self.assertIn('<th>Coefficient 1</th>', result)
        self.assertIn('<th>Coefficient 2</th>', result)
        self.assertIn('<th>Robust covariance</th>', result)
        self.assertIn('<th>Robust correlation</th>', result)
        self.assertIn('<th>Robust t-test</th>', result)
        self.assertIn('<th>Robust p-value</th>', result)

    def test_generate_html_file(self):
        filename = "testfile.html"

        if os.path.exists(filename):
            os.remove(filename)

        generate_html_file(self.estimation_results, filename, overwrite=True)

        # Check that the file was created and contains HTML content
        self.assertTrue(os.path.exists(filename))

        with open(filename, 'r') as file:
            content = file.read()
            self.assertIn('<html>', content)
            self.assertIn('<h1>Estimation report</h1>', content)
            self.assertIn('<h2>Correlation of coefficients</h2>', content)
            self.assertIn('</html>', content)

        # Clean up by removing the file
        os.remove(filename)


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: unary operators

:author: Michel Bierlaire
:date: Sat Sep  9 15:51:53 2023
"""

from __future__ import annotations

import logging

from .base_expressions import Expression, ExpressionOrNumeric
from .convert import validate_and_convert

logger = logging.getLogger(__name__)


class UnaryOperator(Expression):
    """
    Base class for arithmetic expressions that are unary operators.

    Such an expression is the result of the modification of another
    expressions, typically changing its sign.
    """

    def __init__(self, child: ExpressionOrNumeric):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        super().__init__()
        self.child = validate_and_convert(child)
        self.children.append(self.child)



######
# TO DO: combine the validity conditions with the user defined validity
# Do something with the validity message (why it is invalid)


"""Model specification in a multiple expression context

Michel Bierlaire
Sat Jun 28 2025, 12:13:22

Implements a model specification in a multiple expression context (using Catalogs)
"""
from __future__ import annotations

import logging
from typing import Callable

from biogeme.catalog import CentralController, Configuration
from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.parameters import Parameters, get_default_value
from biogeme.results_processing import EstimationResults
from biogeme.tools import ModelNames
from biogeme.validity import Validity
from biogeme_optimization.pareto import SetElement

logger = logging.getLogger(__name__)


class Specification:
    """Implements a specification"""

    database: Database | None = None  #: :class:`biogeme.database.Database` object
    all_results: dict[str:EstimationResults] = {}
    central_controller: CentralController | None = None
    """
        function that generates all the objectives:
        fct(bioResults) -> list[floatNone]
    """
    user_defined_validity_check = None
    """
    function that checks the validity of the results
    """
    generic_name = 'default_name'  #: short name for file names
    biogeme_parameters: Parameters | None = None
    model_names: ModelNames | None = None

    def __init__(
        self,
        configuration: Configuration,
        biogeme_parameters: Parameters | None = None,
    ):
        """Creates a specification from a  configuration

        :param configuration: configuration of the multiple expression
        """
        if not isinstance(configuration, Configuration):
            error_msg = 'Ctor needs an object of type Configuration'
            raise BiogemeError(error_msg)
        self.configuration = configuration
        self.validity = None
        self.biogeme_parameters = biogeme_parameters
        self.maximum_number_parameters = (
            get_default_value(
                name='maximum_number_parameters', section='AssistedSpecification'
            )
            if biogeme_parameters is None
            else biogeme_parameters.get_value(
                name='maximum_number_parameters', section='AssistedSpecification'
            )
        )
        logger.debug(f'{self.maximum_number_parameters=}')

        self._estimate()
        assert (
            self.validity is not None
        ), 'Validity must be set by the _estimate function'

    @classmethod
    def from_string_id(cls, configuration_id: str):
        """Constructor using a configuration"""
        return cls(Configuration.from_string(configuration_id))

    def configure_expression(self) -> None:
        """Configure the expression to the current configuration"""
        self.central_controller.set_configuration(self.configuration)

    @classmethod
    def default_specification(cls) -> Specification:
        """Alternative constructor for generate the default specification"""
        if cls.central_controller is None:
            raise BiogemeError('No central controller has been defined')
        cls.central_controller.reset_selection()
        the_config = cls.central_controller.get_configuration()
        return cls(the_config)

    @property
    def config_id(self) -> str:
        """Defined config_id as a property"""
        return self.configuration.get_string_id()

    @config_id.setter
    def config_id(self, value: str) -> None:
        self.configuration = Configuration.from_string(value)

    def get_results(self) -> EstimationResults:
        """Obtain the estimation results of the specification"""
        the_results = self.all_results.get(self.config_id)
        if the_results is None:
            error_msg = f'No result is available for specification {self.config_id}'
            raise BiogemeError(error_msg)
        return the_results

    def __repr__(self) -> str:
        return str(self.config_id)

    def _estimate(self) -> None:
        """Estimate the parameter of the current specification, if not already done"""
        if self.central_controller is None:
            error_msg = 'No central controller has been provided for the model.'
            raise BiogemeError(error_msg)
        if self.database is None:
            error_msg = 'No database has been provided for the estimation.'
            raise BiogemeError(error_msg)
        if __class__.model_names is None:
            __class__.model_names = ModelNames(prefix=self.generic_name)
        if self.config_id in self.all_results:
            results = self.all_results.get(self.config_id)
        else:
            logger.debug(f'****** Estimate {self.config_id}')
            from biogeme.biogeme import BIOGEME

            the_biogeme: BIOGEME = BIOGEME.from_configuration(
                config_id=self.config_id,
                multiple_expression=self.expression,
                database=self.database,
                generate_html=False,
                generate_yaml=False,
            )
            number_of_parameters = the_biogeme.number_unknown_parameters()
            logger.info(
                f'Model with {number_of_parameters} unknown parameters [max: {self.maximum_number_parameters}]'
            )

            if number_of_parameters > self.maximum_number_parameters:
                logger.info(
                    f'Invalid as it exceeds the maximum number of parameters: {self.maximum_number_parameters}'
                )
                self.validity = Validity(
                    status=False,
                    reason=(
                        f'Too many parameters: {number_of_parameters} > '
                        f'{self.maximum_number_parameters}'
                    ),
                )
                self.all_results[self.config_id] = None
                return

            the_biogeme.model_name = __class__.model_names(self.config_id)
            logger.info(f'*** Estimate {the_biogeme.model_name}')
            results = the_biogeme.quick_estimate()
        self.all_results[self.config_id] = results
        if not results.algorithm_has_converged:
            self.validity = Validity(
                status=False, reason=f'Optimization algorithm has not converged'
            )
            return

        if self.user_defined_validity_check is not None:
            self.validity = self.user_defined_validity_check(results)
        else:
            self.validity = Validity(status=True, reason='')

    def describe(self) -> str:
        """Short description of the solution. Used for reporting.

        :return: short description of the solution.
        :rtype: str
        """
        the_results = self.get_results()
        if the_results is None:
            return f'Invalid model: {self.validity.reason}'
        return f'{the_results.short_summary()}'

    def get_element(
        self, multi_objectives: Callable[[EstimationResults | None], list[float]]
    ) -> SetElement:
        """Obtains the element from the Pareto set corresponding to a specification

        :param multi_objectives: function calculating the objectives
            from the estimation results
        :type multi_objectives: fct(biogeme.results.bioResults) --> list[float]

        :return: element from the Pareto set
        :rtype: biogeme.pareto.SetElement

        """
        the_id = self.config_id
        the_results = self.get_results()
        the_objectives = multi_objectives(the_results)
        element = SetElement(the_id, the_objectives)
        logger.debug(f'{element=}')
        return element



"""
Recycle old pickle files and transform them into YAML files

Michel Bierlaire
Wed Oct 2 06:43:33 2024
"""

import pickle
from datetime import timedelta
from typing import Any

import numpy as np

from .raw_estimation_results import RawEstimationResults, serialize_to_yaml
from ..exceptions import BiogemeError


class RawResults:
    """Class containing the raw results from the estimation"""

    def __init__(
        self,
    ):
        """
        Constructor
        """
        self.modelName = None
        self.userNotes = None
        self.nparam = None
        self.betaValues = None
        self.betaNames = None
        self.initLogLike = None
        self.nullLogLike = None
        self.betas = None
        self.logLike = None
        self.g = None
        self.H = None
        self.bhhh = None
        self.dataname = None
        self.sampleSize = None
        self.numberOfObservations = None
        self.monte_carlo = None
        self.numberOfDraws = None
        self.typesOfDraws = None
        self.excludedData = None
        self.drawsProcessingTime = None
        self.gradientNorm = None
        self.optimizationMessages = None
        self.convergence = None
        self.htmlFileName = None
        self.F12FileName = None
        self.latexFileName = None
        self.pickleFileName = None
        self.bootstrap = None
        self.bootstrap_time = None
        self.secondOrderTable = None


class Beta:
    """Class gathering the information related to the parameters
    of the model
    """

    def __init__(self):
        """
        Constructor
        """
        self.name = None
        self.value = None
        self.lb = None
        self.ub = None
        self.stdErr = None
        self.tTest = None
        self.pValue = None
        self.robust_stdErr = None
        self.robust_tTest = None
        self.robust_pValue = None
        self.bootstrap_stdErr = None
        self.bootstrap_tTest = None
        self.bootstrap_pValue = None


class BiogemeUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        # If the module and class match the missing class, return the fake class
        if module == "biogeme.results" and name == "beta":
            return Beta
        if module == 'biogeme.results' and name == 'rawResults':
            return RawResults
        # Otherwise, proceed as normal
        return super().find_class(module, name)


def read_pickle_biogeme_3_2_14(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    with open(file=filename, mode='br') as file:
        pickled_results = pickle.load(file=file)

    model_name: str = pickled_results.modelName
    user_notes: str = pickled_results.userNotes
    beta_names: list[str] = [beta.name for beta in pickled_results.betas]
    beta_values: list[float] = [beta.value for beta in pickled_results.betas]
    lower_bounds: list[float] = [beta.lb for beta in pickled_results.betas]
    upper_bounds: list[float] = [beta.ub for beta in pickled_results.betas]
    gradient: list[float] = pickled_results.g.tolist()
    hessian: list[list[float]] = pickled_results.H.tolist()
    bhhh: list[list[float]] = pickled_results.bhhh.tolist()
    null_log_likelihood: float = pickled_results.nullLogLike
    initial_log_likelihood: float = pickled_results.initLogLike
    final_log_likelihood: float = pickled_results.logLike
    data_name: str = pickled_results.dataname
    sample_size: int = pickled_results.sampleSize
    number_of_observations: int = pickled_results.numberOfObservations
    monte_carlo: bool = pickled_results.monte_carlo
    number_of_draws: int = pickled_results.numberOfDraws
    types_of_draws: dict[str, str] = pickled_results.typesOfDraws
    number_of_excluded_data: int = pickled_results.excludedData
    draws_processing_time: timedelta = pickled_results.drawsProcessingTime
    optimization_messages: dict[str, Any] = pickled_results.optimizationMessages
    for key, value in optimization_messages.items():
        if isinstance(value, np.ndarray):  # Check if the value is a numpy array
            optimization_messages[key] = value.tolist()
    convergence: bool = pickled_results.convergence
    bootstrap: list[list[float]] = pickled_results.bootstrap.tolist()
    try:
        bootstrap_time: timedelta | None = pickled_results.bootstrap_time
    except AttributeError:
        bootstrap_time = None

    raw_estimation_results = RawEstimationResults(
        model_name=model_name,
        user_notes=user_notes,
        beta_names=beta_names,
        beta_values=beta_values,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        gradient=gradient,
        hessian=hessian,
        bhhh=bhhh,
        null_log_likelihood=null_log_likelihood,
        initial_log_likelihood=initial_log_likelihood,
        final_log_likelihood=final_log_likelihood,
        data_name=data_name,
        sample_size=sample_size,
        number_of_observations=number_of_observations,
        monte_carlo=monte_carlo,
        number_of_draws=number_of_draws,
        types_of_draws=types_of_draws,
        number_of_excluded_data=number_of_excluded_data,
        draws_processing_time=draws_processing_time,
        optimization_messages=optimization_messages,
        convergence=convergence,
        bootstrap=bootstrap,
        bootstrap_time=bootstrap_time,
    )
    return raw_estimation_results


def read_pickle_biogeme_3_2_13(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    with open(file=filename, mode='br') as file:
        pickled_results = BiogemeUnpickler(file=file).load()

    model_name: str = pickled_results.modelName
    user_notes: str = pickled_results.userNotes
    beta_names: list[str] = [beta.name for beta in pickled_results.betas]
    beta_values: list[float] = [beta.value for beta in pickled_results.betas]
    lower_bounds: list[float] = [beta.lb for beta in pickled_results.betas]
    upper_bounds: list[float] = [beta.ub for beta in pickled_results.betas]
    gradient: list[float] = pickled_results.g.tolist()
    hessian: list[list[float]] = pickled_results.H.tolist()
    bhhh: list[list[float]] = pickled_results.bhhh.tolist()
    null_log_likelihood: float = pickled_results.nullLogLike
    initial_log_likelihood: float = pickled_results.initLogLike
    final_log_likelihood: float = pickled_results.logLike
    data_name: str = pickled_results.dataname
    sample_size: int = pickled_results.sampleSize
    number_of_observations: int = pickled_results.numberOfObservations
    monte_carlo: bool = pickled_results.monteCarlo
    number_of_draws: int = pickled_results.numberOfDraws
    types_of_draws: dict[str, str] = pickled_results.typesOfDraws
    number_of_excluded_data: int = pickled_results.excludedData
    draws_processing_time: timedelta = pickled_results.drawsProcessingTime
    optimization_messages: dict[str, Any] = pickled_results.optimizationMessages
    for key, value in optimization_messages.items():
        if isinstance(value, np.ndarray):  # Check if the value is a numpy array
            optimization_messages[key] = value.tolist()
    convergence: bool = pickled_results.convergence
    bootstrap: list[list[float]] = pickled_results.bootstrap.tolist()
    try:
        bootstrap_time: timedelta | None = pickled_results.bootstrap_time
    except AttributeError:
        bootstrap_time = None

    raw_estimation_results = RawEstimationResults(
        model_name=model_name,
        user_notes=user_notes,
        beta_names=beta_names,
        beta_values=beta_values,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        gradient=gradient,
        hessian=hessian,
        bhhh=bhhh,
        null_log_likelihood=null_log_likelihood,
        initial_log_likelihood=initial_log_likelihood,
        final_log_likelihood=final_log_likelihood,
        data_name=data_name,
        sample_size=sample_size,
        number_of_observations=number_of_observations,
        monte_carlo=monte_carlo,
        number_of_draws=number_of_draws,
        types_of_draws=types_of_draws,
        number_of_excluded_data=number_of_excluded_data,
        draws_processing_time=draws_processing_time,
        optimization_messages=optimization_messages,
        convergence=convergence,
        bootstrap=bootstrap,
        bootstrap_time=bootstrap_time,
    )
    return raw_estimation_results


def read_pickle_biogeme_3_2_12(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    with open(file=filename, mode='br') as file:
        pickled_results = BiogemeUnpickler(file=file).load()

    model_name: str = pickled_results.modelName
    user_notes: str = pickled_results.userNotes
    beta_names: list[str] = [beta.name for beta in pickled_results.betas]
    beta_values: list[float] = [beta.value for beta in pickled_results.betas]
    lower_bounds: list[float] = [beta.lb for beta in pickled_results.betas]
    upper_bounds: list[float] = [beta.ub for beta in pickled_results.betas]
    gradient: list[float] = pickled_results.g.tolist()
    hessian: list[list[float]] = pickled_results.H.tolist()
    bhhh: list[list[float]] = pickled_results.bhhh.tolist()
    null_log_likelihood: float = pickled_results.nullLogLike
    initial_log_likelihood: float = pickled_results.initLogLike
    final_log_likelihood: float = pickled_results.logLike
    data_name: str = pickled_results.dataname
    sample_size: int = pickled_results.sampleSize
    number_of_observations: int = pickled_results.numberOfObservations
    monte_carlo: bool = pickled_results.monteCarlo
    number_of_draws: int = pickled_results.numberOfDraws
    types_of_draws: dict[str, str] = pickled_results.typesOfDraws
    number_of_excluded_data: int = pickled_results.excludedData
    draws_processing_time: timedelta = pickled_results.drawsProcessingTime
    optimization_messages: dict[str, Any] = pickled_results.optimizationMessages
    for key, value in optimization_messages.items():
        print(f'{key} {type(value)}')
        if isinstance(
            value, (np.ndarray, np.float64, np.float32)
        ):  # Check if the value is a numpy array
            optimization_messages[key] = value.tolist()
    convergence: bool = True
    bootstrap: list[list[float]] = (
        pickled_results.bootstrap.tolist()
        if pickled_results.bootstrap is not None
        else None
    )
    try:
        bootstrap_time: timedelta | None = pickled_results.bootstrap_time
    except AttributeError:
        bootstrap_time = None

    raw_estimation_results = RawEstimationResults(
        model_name=model_name,
        user_notes=user_notes,
        beta_names=beta_names,
        beta_values=beta_values,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        gradient=gradient,
        hessian=hessian,
        bhhh=bhhh,
        null_log_likelihood=null_log_likelihood,
        initial_log_likelihood=initial_log_likelihood,
        final_log_likelihood=final_log_likelihood,
        data_name=data_name,
        sample_size=sample_size,
        number_of_observations=number_of_observations,
        monte_carlo=monte_carlo,
        number_of_draws=number_of_draws,
        types_of_draws=types_of_draws,
        number_of_excluded_data=number_of_excluded_data,
        draws_processing_time=draws_processing_time,
        optimization_messages=optimization_messages,
        convergence=convergence,
        bootstrap=bootstrap,
        bootstrap_time=bootstrap_time,
    )
    return raw_estimation_results


def read_pickle_biogeme_3_2_11(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    return read_pickle_biogeme_3_2_12(filename=filename)


def read_pickle_biogeme_3_2_10(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    return read_pickle_biogeme_3_2_12(filename=filename)


def read_pickle_biogeme_3_2_8(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    return read_pickle_biogeme_3_2_12(filename=filename)


def read_pickle_biogeme_3_2_7(filename: str) -> RawEstimationResults:
    """

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    return read_pickle_biogeme_3_2_12(filename=filename)


def read_pickle_biogeme(filename: str) -> RawEstimationResults:
    """Read an old pickle file, when the version of Biogeme used to create it is unknown

    :param filename: name of the pickle file
    :return: raw estimation results
    """
    try:
        results = read_pickle_biogeme_3_2_14(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_13(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_12(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_11(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_10(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_8(filename=filename)
        return results
    except AttributeError:
        ...

    try:
        results = read_pickle_biogeme_3_2_7(filename=filename)
        return results
    except AttributeError:
        ...

    error_msg = f'It was not possible to identify the format of the file {filename}'
    raise BiogemeError(error_msg)


def pickle_to_yaml(pickle_filename: str, yaml_filename: str) -> None:
    """
    Transforms a pickle file into a YAML format with the estimation results.

    :param pickle_filename: name of the input file
    :param yaml_filename: name of the output file
    """
    results = read_pickle_biogeme(filename=pickle_filename)
    serialize_to_yaml(data=results, filename=yaml_filename)



"""

Arc elasticities
================

We use a previously estimated nested logit model and calculate arc elasticities.

Details about this example are available in Section 3 of `Bierlaire (2018)
Calculating indicators with PandasBiogeme
<http://transp-or.epfl.ch/documents/technicalReports/Bier18a.pdf>`_

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:00:10 2023

"""

import sys
import biogeme.biogeme as bio
import biogeme.exceptions as excep
from biogeme import models
import biogeme.results as res
from optima_data import database, normalized_weight
from scenarios import scenario

# %%
# Obtain the specification for the default scenario
# The definition of the scenarios is available in :ref:`scenarios`.
V, nests, _, MarginalCostPT = scenario()

# %%
# Obtain the expression for the choice probability of the public transportation.
prob_PT = models.nested(V, None, nests, 0)

# %%
# We investigate a scenario where the price for public transportation
# increases by 20%. We extract the corresponding scenario.
V_after, _, _, MarginalCostPT_after = scenario(factor=1.2)
prob_PT_after = models.nested(V_after, None, nests, 0)

# %%
# Disaggregate elasticities
direct_elas_pt = (
    (prob_PT_after - prob_PT)
    * MarginalCostPT
    / (prob_PT * (MarginalCostPT_after - MarginalCostPT))
)

# %%
# Formulas to simulate.
simulate = {
    'weight': normalized_weight,
    'Prob. PT': prob_PT,
    'direct_elas_pt': direct_elas_pt,
}

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, simulate)

# %%
# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='saved_results/b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02estimation.py in order to generate '
        'the file b02estimation.pickle.'
    )

# %%
# `simulated_values` is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulated_values = the_biogeme.simulate(results.getBetaValues())
simulated_values

# %%
# We calculate the aggregate elasticities.

# %%
# First, the weighted probabilities.
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. PT']
)

# %%
# Then the denominator of the aggregate elasticity expression.
denominator_pt = simulated_values['Weighted prob. PT'].sum()

# %%
# And finally the aggregate elasticities themselves.
direct_elas_pt = (
    simulated_values['Weighted prob. PT']
    * simulated_values['direct_elas_pt']
    / denominator_pt
).sum()

print(
    f'Aggregate direct arc elasticity of public transportation wrt cost: '
    f'{direct_elas_pt:.3g}'
)



import unittest

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.data.swissmetro import (
    CHOICE,
    GA,
    PURPOSE,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, Elem, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)


B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

tau1 = Beta('tau1', -1, None, 0, 0)
delta2 = Beta('delta2', 2, 0, None, 0)

tau2 = tau1 + delta2


TRAIN_COST = TRAIN_CO * (GA == 0)


#  Utility

U = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED


ChoiceProba = {
    1: 1 - dist.logisticcdf(U - tau1),
    2: dist.logisticcdf(U - tau1) - dist.logisticcdf(U - tau2),
    3: dist.logisticcdf(U - tau2),
}

logprob = log(Elem(ChoiceProba, CHOICE))


class test_18(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=None,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -5789.309, 2)


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import numpy as np
from biogeme.validation.prepare_validation import split


class TestSplitFunction(unittest.TestCase):
    def setUp(self):
        # Create a simple DataFrame for tests
        self.df = pd.DataFrame(
            {
                'id': range(100),
                'group': [i // 10 for i in range(100)],
                'feature': np.random.rand(100),
            }
        )

    def test_basic_split(self):
        result = split(self.df, slices=5)
        self.assertEqual(len(result), 5)
        total_rows = sum(len(ev.estimation) + len(ev.validation) for ev in result)
        self.assertEqual(total_rows, 5 * len(self.df))
        for ev in result:
            self.assertEqual(len(ev.estimation) + len(ev.validation), len(self.df))

    def test_split_with_groups(self):
        result = split(self.df, slices=5, groups='group')
        self.assertEqual(len(result), 5)
        all_ids = set(self.df['id'])
        for ev in result:
            group_ids_val = set(self.df.loc[ev.validation, 'group'].unique())
            group_ids_est = set(self.df.loc[ev.estimation, 'group'].unique())
            self.assertTrue(group_ids_val.isdisjoint(group_ids_est))
            combined_ids = set(ev.estimation).union(set(ev.validation))
            self.assertSetEqual(combined_ids, set(self.df.index))

    def test_invalid_slices(self):
        with self.assertRaises(ValueError):
            split(self.df, slices=1)

    def test_split_preserves_all_data(self):
        result = split(self.df, slices=4)
        combined_ids = pd.concat([self.df.loc[ev.validation] for ev in result])['id']
        self.assertSetEqual(set(combined_ids), set(self.df['id']))


if __name__ == '__main__':
    unittest.main()



import unittest

import pandas as pd

from biogeme.database import Database
from biogeme.expressions import Beta, Numeric, Variable
from biogeme.mdcev import GammaProfile, Translated, Generalized, NonMonotonic


class TestMdcev(unittest.TestCase):

    def setUp(self):
        first_row = {
            'PersonID': 1.0,
            'weight': 0.2244021,
            'hhsize': 1.0,
            'childnum': 0.0,
            'faminc': 6.0,
            'faminc25K': 1.0,
            'income': 17.5,
            'employed': 0.0,
            'fulltime': 0.0,
            'spousepr': 0.0,
            'spousemp': 0.0,
            'male': 0.0,
            'married': 0.0,
            'age': 85.0,
            'age2': 7225.0,
            'age15_40': 0.0,
            'age41_60': 0.0,
            'age61_85': 1.0,
            'bachigher': 0.0,
            'white': 0.0,
            'metro': 1.0,
            'diaryday': 1.0,
            'Sunday': 1.0,
            'holiday': 0.0,
            'weekearn': -1.0,
            'weekwordur': -1.0,
            'hhchild': 0.0,
            'ohhchild': 0.0,
            't1': 0.0,
            't2': 240.0,
            't3': 0.0,
            't4': 30.0,
            'number_chosen': 2.0,
        }
        self.one_row = Database('one_row', pd.DataFrame([first_row]))
        weight = Variable('orig_weight') * 1.7718243289995812

        # %
        # Parameters to be estimated
        cte_shopping = Beta('cte_shopping', -0.736, None, None, 0)
        cte_socializing = Beta('cte_socializing', -0.736, None, None, 0)
        cte_recreation = Beta('cte_recreation', -0.837, None, None, 0)

        number_members_socializing = Beta(
            'number_members_socializing', 0.0166, None, None, 0
        )
        number_members_recreation = Beta(
            'number_members_recreation', 0.0198, None, None, 0
        )

        metropolitan_shopping = Beta('metropolitan_shopping', 0.062, None, None, 0)

        male_shopping = Beta('male_shopping', 0.0979, None, None, 0)
        male_socializing = Beta('male_socializing', 0.127, None, None, 0)
        male_recreation = Beta('male_recreation', 0.219, None, None, 0)

        age_15_40_shopping = Beta('age_15_40_shopping', 0.0755, None, None, 0)
        age_15_40_recreation = Beta('age_15_40_recreation', 0.121, None, None, 0)

        age_41_60_socializing = Beta('age_41_60_socializing', -0.0646, None, None, 0)
        age_41_60_personal = Beta('age_41_60_personal', -0.0551, None, None, 0)

        bachelor_socializing = Beta('bachelor_socializing', -0.051, None, None, 0)
        bachelor_personal = Beta('bachelor_personal', -0.0455, None, None, 0)

        white_personal = Beta('white_personal', -0.0884, None, None, 0)

        spouse_shopping = Beta('spouse_shopping', 0.0628, None, None, 0)
        spouse_recreation = Beta('spouse_recreation', -0.0656, None, None, 0)

        employed_shopping = Beta('employed_shopping', 0.0462, None, None, 0)

        sunday_socializing = Beta('sunday_socializing', 0.0963, None, None, 0)
        sunday_personal = Beta('sunday_personal', 0.0907, None, None, 0)

        # %
        # Definition of the utility functions
        shopping = (
            cte_shopping
            + metropolitan_shopping * Variable('metro')
            + male_shopping * Variable('male')
            + age_15_40_shopping * Variable('age15_40')
            + spouse_shopping * Variable('spousepr')
            + employed_shopping * Variable('employed')
        )

        socializing = (
            cte_socializing
            + number_members_socializing * Variable('hhsize')
            + male_socializing * Variable('male')
            + age_41_60_socializing * Variable('age41_60')
            + bachelor_socializing * Variable('bachigher')
            + sunday_socializing * Variable('Sunday')
        )

        recreation = (
            cte_recreation
            + number_members_recreation * Variable('hhsize')
            + male_recreation * Variable('male')
            + age_15_40_recreation * Variable('age15_40')
            + spouse_recreation * Variable('spousepr')
        )

        personal = (
            age_41_60_personal * Variable('age41_60')
            + bachelor_personal * Variable('bachigher')
            + white_personal * Variable('white')
            + sunday_personal * Variable('Sunday')
        )

        baseline_utilities = {
            1: shopping,
            2: socializing,
            3: recreation,
            4: personal,
        }

        # Observed consumed quantities
        consumed_quantities = {
            1: Variable('t1') / 60.0,
            2: Variable('t2') / 60.0,
            3: Variable('t3') / 60.0,
            4: Variable('t4') / 60.0,
        }
        gamma_shopping = Beta('gamma_shopping', 3.49, 0.001, None, 0)
        gamma_socializing = Beta('gamma_socializing', 11.1, 0.001, None, 0)
        gamma_recreation = Beta('gamma_recreation', 15.4, 0.001, None, 0)
        gamma_personal = Beta('gamma_personal', 1.56, 0.001, None, 0)

        scale_parameter = Beta('scale', 3.83, 0.0001, None, 0)

        gamma_parameters = {
            1: gamma_shopping,
            2: gamma_socializing,
            3: gamma_recreation,
            4: gamma_personal,
        }

        alpha_parameter = {
            1: Beta('alpha_1', 0.5, 0, 1, 0),
            2: Beta('alpha_1', 0.5, 0, 1, 0),
            3: Beta('alpha_1', 0.5, 0, 1, 0),
            4: Beta('alpha_1', 0.5, 0, 1, 0),
        }

        prices = {1: Numeric(1), 2: Numeric(1), 3: Numeric(1), 4: Numeric(1)}

        self.the_gamma_profile_with_scale = GammaProfile(
            model_name='gamma_profile_with_scale',
            baseline_utilities=baseline_utilities,
            prices=prices,
            gamma_parameters=gamma_parameters,
            scale_parameter=scale_parameter,
            weights=weight,
        )

        self.the_translated = Translated(
            model_name='translated',
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameter,
            scale_parameter=scale_parameter,
            weights=weight,
        )

        self.generalized = Generalized(
            model_name='generalized',
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameter,
            scale_parameter=scale_parameter,
            prices=prices,
            weights=weight,
        )

        self.non_monotonic = NonMonotonic(
            model_name='non_monotonic',
            baseline_utilities=baseline_utilities,
            mu_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameter,
            scale_parameter=scale_parameter,
            weights=weight,
        )

    def test_gamma_profile(self):
        error_msg = self.the_gamma_profile_with_scale.validation(one_row=self.one_row)
        self.assertEqual(error_msg, [], '\n'.join(error_msg))

    def test_translated(self):
        error_msg = self.the_translated.validation(one_row=self.one_row)
        self.assertEqual(error_msg, [], '\n'.join(error_msg))

    def test_generalized(self):
        error_msg = self.generalized.validation(one_row=self.one_row)
        self.assertEqual(error_msg, [], '\n'.join(error_msg))

    def test_non_monotonic(self):
        error_msg = self.non_monotonic.validation(one_row=self.one_row)
        self.assertEqual(error_msg, [], '\n'.join(error_msg))



""" Tests for checks.py

Michel Bierlaire
Fri Apr 5 11:21:26 2024
"""

import unittest

from biogeme.tools.checks import check_consistency_of_named_dicts


class TestCheckConsistencyOfNamedDicts(unittest.TestCase):
    def test_all_dictionaries_consistent(self):
        """Test case where all dictionaries have the same set of keys."""
        dicts = {
            "first": {1: 'a', 2: 'b'},
            "second": {1: 'c', 2: 'd'},
            "third": {1: 'e', 2: 'f'},
        }
        consistent, report = check_consistency_of_named_dicts(dicts)
        self.assertTrue(consistent)
        self.assertIsNone(report)

    def test_dictionaries_with_inconsistencies(self):
        """Test case where dictionaries have different sets of keys."""
        dicts = {
            "first": {1: 'a', 2: 'b'},
            "second": {1: 'c', 3: 'd'},
            "third": {2: 'e', 3: 'f'},
        }
        consistent, report = check_consistency_of_named_dicts(dicts)
        self.assertFalse(consistent)
        self.assertIsNotNone(report)
        self.assertIn("second missing keys compared to first:", report)
        self.assertIn("second has extra keys compared to first:", report)

    def test_empty_dictionaries_consistent(self):
        """Test case where all dictionaries are empty."""
        dicts = {"first": {}, "second": {}, "third": {}}
        consistent, report = check_consistency_of_named_dicts(dicts)
        self.assertTrue(consistent)
        self.assertIsNone(report)

    def test_one_empty_dictionary_inconsistent(self):
        """Test case where one dictionary is empty and others are not."""
        dicts = {"first": {1: 'a'}, "second": {}, "third": {1: 'b'}}
        consistent, report = check_consistency_of_named_dicts(dicts)
        self.assertFalse(consistent)
        self.assertIsNotNone(report)
        self.assertIn("second missing keys compared to first:", report)

    def test_single_dictionary_consistent(self):
        """Test case with only one dictionary."""
        dicts = {"first": {1: 'a', 2: 'b'}}
        consistent, report = check_consistency_of_named_dicts(dicts)
        self.assertTrue(consistent)
        self.assertIsNone(report)


if __name__ == '__main__':
    unittest.main()



"""Defines exceptions for Biogeme

:author: Michel Bierlaire

:date: Fri Dec  8 20:34:09 2023

"""


class BiogemeError(Exception):
    """Defines a generic exception for Biogeme."""


class FileNotFound(BiogemeError):
    """Value of out range error."""


class ValueOutOfRange(BiogemeError):
    """Value of out range error."""


class DuplicateError(BiogemeError):
    """duplicate error."""


class NotImplementedError(BiogemeError):
    """Not implemented error."""



"""Tests for nary expressions

Michel Bierlaire
Wed Mar 26 11:46:06 2025
"""

import unittest
import warnings

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Beta,
    ConditionalSum,
    ConditionalTermTuple,
    Elem,
    LinearTermTuple,
    LinearUtility,
    MultipleSum,
    Numeric,
    Variable,
    bioLinearUtility,
    bioMultSum,
)


class TestNaryExpressions(unittest.TestCase):
    def test_conditional_sum_value(self):
        terms = [
            ConditionalTermTuple(Numeric(1), Numeric(10)),
            ConditionalTermTuple(Numeric(0), Numeric(100)),
            ConditionalTermTuple(Numeric(2), Numeric(5)),
        ]
        expr = ConditionalSum(terms)
        self.assertEqual(expr.get_value(), 15.0)

    def test_multiple_sum_value(self):
        expr = MultipleSum([Numeric(1), Numeric(2), Numeric(3)])
        self.assertEqual(expr.get_value(), 6.0)

    def test_elem_valid_key(self):
        mapping = {0: Numeric(1), 1: Numeric(2)}
        expr = Elem(mapping, Numeric(1))
        self.assertEqual(expr.get_value(), 2.0)

    def test_elem_invalid_key(self):
        mapping = {0: Numeric(1)}
        expr = Elem(mapping, Numeric(3))
        with self.assertRaises(BiogemeError):
            expr.get_value()

    def test_linear_utility_value(self):
        b = Beta("beta", 2.0, None, None, 0)
        x = Variable("x")
        x.variable_id = 0
        expr = LinearUtility([LinearTermTuple(b, x)])
        self.assertIn("LinearUtility", repr(expr))

    def test_deprecated_bioMultSum(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = bioMultSum([Numeric(1)])
            self.assertTrue(
                any(issubclass(wi.category, DeprecationWarning) for wi in w)
            )

    def test_deprecated_bioLinearUtility(self):
        b = Beta("b", 1.0, None, None, 0)
        x = Variable("x")
        x.variable_id = 0
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = bioLinearUtility([LinearTermTuple(b, x)])
            self.assertTrue(
                any(issubclass(wi.category, DeprecationWarning) for wi in w)
            )


if __name__ == '__main__':
    unittest.main()



"""Unit tests for the formula implementation

Michel Bierlaire
Sat Mar 29 10:56:48 2025
"""

import unittest

import numpy as np
import pandas as pd

from biogeme.calculator import CompiledFormulaEvaluator, calculate_single_formula
from biogeme.calculator.single_formula import evaluate_model_per_row
from biogeme.database import Database
from biogeme.expressions import Beta, Variable
from biogeme.model_elements import ModelElements
from biogeme.second_derivatives import SecondDerivativesMode


class TestCompiledFormulaEvaluator(unittest.TestCase):
    def setUp(self):
        # Simple synthetic dataset
        self.data = pd.DataFrame.from_dict({'x': [1.0, 2.0, 3.0]})
        self.database = Database('test', self.data)
        self.x = Variable('x')
        self.beta = Beta('beta', 1.0, None, None, 0)
        self.expression = self.beta * self.x
        self.betas = {'beta': 2.0}
        self.draws = None

    def test_function_only(self):
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        evaluator = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        output = evaluator.evaluate(
            self.betas, gradient=False, hessian=False, bhhh=False
        )
        expected_value = float(np.sum(np.array(self.data['x']) * self.betas['beta']))
        self.assertAlmostEqual(output.function, expected_value, places=6)
        self.assertIsNone(output.gradient)
        self.assertIsNone(output.hessian)
        self.assertIsNone(output.bhhh)

    def test_function_and_gradient(self):
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        evaluator = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        output = evaluator.evaluate(
            self.betas, gradient=True, hessian=False, bhhh=False
        )
        self.assertIsNotNone(output.gradient)
        expected_gradient = np.sum(self.data['x'])
        self.assertAlmostEqual(output.gradient[0], expected_gradient, places=6)

    def test_function_gradient_hessian(self):
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        evaluator = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        output = evaluator.evaluate(self.betas, gradient=True, hessian=True, bhhh=False)
        self.assertIsNotNone(output.hessian)
        expected_hessian = np.zeros((1, 1))
        self.assertTrue(np.allclose(output.hessian, expected_hessian))

    def test_function_bhhh(self):
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        evaluator = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )

        output = evaluator.evaluate(self.betas, gradient=True, hessian=False, bhhh=True)
        self.assertIsNotNone(output.bhhh)
        self.assertEqual(output.bhhh.shape, (1, 1))

    def test_missing_beta_uses_default(self):
        betas_missing = {}
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        evaluator = CompiledFormulaEvaluator(
            model_elements=model_elements,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        output = evaluator.evaluate(
            betas_missing, gradient=False, hessian=False, bhhh=False
        )
        expected_value = float(
            np.sum(np.array(self.data['x']) * 1.0)
        )  # default beta = 1.0
        self.assertAlmostEqual(output.function, expected_value, places=6)

    def test_legacy_calculate_single_formula(self):
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        output = calculate_single_formula(
            model_elements=model_elements,
            the_betas=self.betas,
            gradient=True,
            hessian=False,
            bhhh=False,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        self.assertIsNotNone(output.function)
        self.assertIsNotNone(output.gradient)


class TestEvaluateExpressionPerRow(unittest.TestCase):
    def setUp(self):
        # Sample data: x = [1.0, 2.0, 3.0]
        self.data = pd.DataFrame({'x': [1.0, 2.0, 3.0]})
        self.database = Database('test', self.data)
        self.draws = None

        # Expression: beta * x
        self.x = Variable('x')
        self.beta = Beta('beta', 1.0, None, None, 0)
        self.expression = self.beta * self.x

    def test_evaluate_expression_per_row_with_specified_beta(self):
        betas = {'beta': 2.0}
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )

        results = evaluate_model_per_row(
            model_elements=model_elements,
            the_betas=betas,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected = np.array([2.0, 4.0, 6.0])
        np.testing.assert_array_almost_equal(results, expected)

    def test_evaluate_expression_per_row_with_default_beta(self):
        # No beta passed, should use default (1.0)
        betas = {}
        model_elements = ModelElements.from_expression_and_weight(
            log_like=self.expression, weight=None, database=self.database
        )
        results = evaluate_model_per_row(
            model_elements=model_elements,
            the_betas=betas,
            second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
            numerically_safe=False,
        )
        expected = np.array([1.0, 2.0, 3.0])
        np.testing.assert_array_almost_equal(results, expected)


if __name__ == '__main__':
    unittest.main()



"""Arithmetic expressions accepted by Biogeme: nary operators

Michel Bierlaire
Sat Sep  9 15:29:36 2023
"""

from __future__ import annotations

import logging
from typing import NamedTuple, TYPE_CHECKING

import jax.numpy as jnp
from biogeme.exceptions import BiogemeError

from .base_expressions import Expression
from .beta_parameters import Beta
from .elementary_expressions import Elementary, TypeOfElementaryExpression
from .jax_utils import JaxFunctionType
from .variable import Variable

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class LinearTermTuple(NamedTuple):
    beta: Beta
    x: Variable


class LinearUtility(Expression):
    """When the utility function is linear, it is expressed as a list of
    terms, where a parameter multiplies a variable.
    """

    def __init__(self, list_of_terms: list[LinearTermTuple]):
        """Constructor

        :param list_of_terms: a list of tuple. Each tuple contains first
             a Beta parameter, second the name of a variable.
        :type list_of_terms: list(biogeme.expressions.Expression,
            biogeme.expressions.Expression)

        :raises biogeme.exceptions.BiogemeError: if the object is not
                        a list of tuples (parameter, variable)

        """
        super().__init__()

        the_error = ''
        first = True

        for b, v in list_of_terms:
            if not isinstance(b, Beta) or not isinstance(v, Variable):
                raise BiogemeError(
                    f'Each term must be a (Beta, Variable) pair. Got: ({b}, {v})'
                )

        if not first:
            raise BiogemeError(the_error)

        self.betas, self.variables = zip(*list_of_terms)

        self.betas = list(self.betas)  #: list of parameters

        self.variables = list(self.variables)  #: list of variables

        self.list_of_terms = list(zip(self.betas, self.variables))
        """ List of terms """

        self.children += self.betas + self.variables

    def deep_flat_copy(self) -> LinearUtility:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_list_of_terms = [
            LinearTermTuple(beta=term[0].deep_flat_copy(), x=term[1].deep_flat_copy())
            for term in self.list_of_terms
        ]
        return type(self)(list_of_terms=copy_list_of_terms)

    def __str__(self) -> str:
        return ' + '.join([f'{b} * {x}' for b, x in self.list_of_terms])

    def __repr__(self) -> str:
        return f"LinearUtility({repr(self.list_of_terms)})"

    def dict_of_elementary_expression(
        self, the_type: TypeOfElementaryExpression
    ) -> dict[str, Elementary]:
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression

        :return: returns a dict with the variables appearing in the
               expression the keys being their names.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        if the_type == TypeOfElementaryExpression.BETA:
            return {x.name: x for x in self.betas}

        if the_type == TypeOfElementaryExpression.FREE_BETA:
            return {x.name: x for x in self.betas if x.status == 0}

        if the_type == TypeOfElementaryExpression.FIXED_BETA:
            return {x.name: x for x in self.betas if x.status != 0}

        if the_type == TypeOfElementaryExpression.VARIABLE:
            return {x.name: x for x in self.variables}

        return {}

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        beta_fns = [
            b.recursive_construct_jax_function(numerically_safe=numerically_safe)
            for b in self.betas
        ]
        variable_fns = [
            v.recursive_construct_jax_function(numerically_safe=numerically_safe)
            for v in self.variables
        ]

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            beta_values = jnp.array(
                [
                    fn(parameters, one_row, the_draws, the_random_variables)
                    for fn in beta_fns
                ]
            )
            variable_values = jnp.array(
                [
                    fn(parameters, one_row, the_draws, the_random_variables)
                    for fn in variable_fns
                ]
            )
            return jnp.dot(beta_values, variable_values)

        return the_jax_function



"""

Catalog for alternative specific coefficients
=============================================

Investigate alternative specific parameters:

    - two specifications for the travel time coefficient: generic, and
      alternative specific,
    - two specifications for the travel cost coefficient: generic, and
      alternative specific,

for a total of 4 specifications.
See `Bierlaire and Ortelli (2023)
<https://transp-or.epfl.ch/documents/technicalReports/BierOrte23.pdf>`_.


:author: Michel Bierlaire, EPFL
:date: Thu Jul 13 16:18:10 2023

"""
import biogeme.biogeme as bio
import biogeme.biogeme_logging as blog
from biogeme import models
from biogeme.expressions import Beta
from biogeme.catalog import generic_alt_specific_catalogs
from biogeme.results import compile_estimation_results, pareto_optimal


# %%
# See :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Catalog for travel time coefficient.
(B_TIME_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='B_TIME', beta_parameters=[B_TIME], alternatives=('TRAIN', 'SM', 'CAR')
)

# %%
# Catalog for travel cost coefficient.
(B_COST_catalog_dict,) = generic_alt_specific_catalogs(
    generic_name='B_COST', beta_parameters=[B_COST], alternatives=('TRAIN', 'SM', 'CAR')
)

# %%
# Definition of the utility functions.
V1 = (
    ASC_TRAIN
    + B_TIME_catalog_dict['TRAIN'] * TRAIN_TT_SCALED
    + B_COST_catalog_dict['TRAIN'] * TRAIN_COST_SCALED
)
V2 = (
    B_TIME_catalog_dict['SM'] * SM_TT_SCALED
    + B_COST_catalog_dict['SM'] * SM_COST_SCALED
)
V3 = (
    ASC_CAR
    + B_TIME_catalog_dict['CAR'] * CAR_TT_SCALED
    + B_COST_catalog_dict['CAR'] * CAR_CO_SCALED
)

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b01alt_spec'
the_biogeme.generate_html = False
the_biogeme.generate_pickle = False

# %%
# Estimate the parameters.
dict_of_results = the_biogeme.estimate_catalog()

# %%
# Number of estimated models.
print(f'A total of {len(dict_of_results)} models have been estimated')

# %%
# All estimation results
compiled_results, specs = compile_estimation_results(
    dict_of_results, use_short_names=True
)

# %%
compiled_results

# %%
# Glossary
for short_name, spec in specs.items():
    print(f'{short_name}\t{spec}')

# %%
# Estimation results of the Pareto optimal models.
pareto_results = pareto_optimal(dict_of_results)
compiled_pareto_results, pareto_specs = compile_estimation_results(
    pareto_results, use_short_names=True
)

# %%
compiled_pareto_results

# %%
# Glossary.
for short_name, spec in pareto_specs.items():
    print(f'{short_name}\t{spec}')



"""

biogeme.expressions
===================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Tue Nov 21 14:53:49 2023
"""

from biogeme.version import getText
import numpy as np
import pandas as pd
import biogeme.expressions as ex
import biogeme.database as db
import biogeme.exceptions as excep
from biogeme import models
from biogeme import tools
from biogeme.expressions import IdManager, TypeOfElementaryExpression
import biogeme.biogeme_logging as blog

# %%
# Version of Biogeme.
print(getText())

# %%
logger = blog.get_screen_logger(level=blog.INFO)

# %%
# Simple expressions
# ------------------
# Simple expressions can be evaluated both with the functions
# `getValue`(implemented in Python) and the `getValue_c` (implemented
# in C++). They do not require a database.

# %%
x = ex.Beta('x', 2, None, None, 1)
x

# %%
x.getValue()

# %%
x.getValue_c(prepareIds=True)


# %%
y = ex.Beta('y', 3, None, None, 1)
y

# %%
y.getValue()

# %%
y.getValue_c(prepareIds=True)

# %%
# Note that if the parameter has to be estimated, its value cannot be obtained.
unknown_parameter = ex.Beta('x', 2, None, None, 0)
try:
    unknown_parameter.getValue()
except excep.BiogemeError as e:
    print(e)

# %%
one = ex.Numeric(1)
one

# %%
one.getValue()

# %%
one.getValue_c(prepareIds=True)

# %%
# Addition
z = x + y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Substraction
z = x - y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Multiplication
z = x * y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Division
z = x / y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Power
z = x**y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Exponential
z = ex.exp(x)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Logarithm
z = ex.log(x)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Minimum
z = ex.bioMin(x, y)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Maximum
z = ex.bioMax(x, y)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# And
z = x & y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
z = x & 0
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Or

# %%
z = x | y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
z = ex.Numeric(0) | ex.Numeric(0)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Equal
z = x == y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
z = (x + 1) == y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Not equal
z = x != y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
z = (x + 1) != y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Lesser or equal
z = x <= y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Greater or equal
z = x >= y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Lesser than
z = x < y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Greater than
z = x > y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Opposite
z = -x
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Sum of multiples expressions
listOfExpressions = [x, y, 1 + x, 1 + y]
z = ex.bioMultSum(listOfExpressions)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# The result is the same as the following, but it implements the sum
# in a more efficient way.
z = x + y + 1 + x + 1 + y
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
# Element: this expression considers a dictionary of expressions, and
# an expression for the index. The index is evaluated, and the value
# of the corresponding expression in the dictionary is returned.
my_dict = {1: ex.exp(-1), 2: ex.log(1.2), 3: 1234}

# %%
index = x
index.getValue()

# %%
z = ex.Elem(my_dict, index)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
index = x - 1
index.getValue()

# %%
z = ex.Elem(my_dict, index)
z.getValue()

# %%
z.getValue_c(prepareIds=True)

# %%
index = x - 2
index.getValue()

# %%
# If the value returned as index does not corresponds to an entry in
# the dictionary, an exception is raised.

# %%
z = ex.Elem(my_dict, index)
try:
    z.getValue()
except excep.BiogemeError as e:
    print(f'Exception raised: {e}')

# %%
z = ex.Elem(my_dict, index)
try:
    z.getValue_c(prepareIds=True)
except RuntimeError as e:
    print(f'Exception raised: {e}')

# %%
# Complex expressions
# -------------------
# When an expression is deemed complex in Biogeme, the `getValue`
# function is not available. Only the `getValue_c` function must be
# used. It calculates the expressions using a C++ implementation of
# the expression.

# %%
# Normal CDF: it calculates
#
# .. math:: \Phi(x) = \frac{1}{\sqrt{2\pi}}\int_{-\infty}^{x}
#     e^{-\frac{1}{2}\omega^2}d\omega.
z = ex.bioNormalCdf(x)
z.getValue_c(prepareIds=True)

# %%
z = ex.bioNormalCdf(0)
z.getValue_c(prepareIds=True)

# %%
# Derivative
z = 30 * x + 20 * y

# %%
zx = ex.Derive(z, 'x')
zx.getValue_c(prepareIds=True)

# %%
zx = ex.Derive(z, 'y')
zx.getValue_c(prepareIds=True)

# %%
# Integral: let's calculate the integral of the pdf of a normal distribution:
#
# .. math:: \frac{1}{\sqrt{2\pi}}\int_{-\infty}^{+\infty}
#     e^{-\frac{1}{2}\omega^2}d\omega = 1.
omega = ex.RandomVariable('omega')
pdf = ex.exp(-omega * omega / 2)
z = ex.Integrate(pdf, 'omega') / np.sqrt(2 * np.pi)
z.getValue_c(prepareIds=True)

# %%
# In order to change the bounds of integration, a change of variables
# must be performed. Let's calculate
#
# .. math:: \int_0^1 x^2 dx=\frac{1}{3}.
#
# If :math:`a` is the lower bound of integration, and :math:`b` is the upper
# bound, the change of variable is
#
# .. math:: x = a + \frac{b-a}{1+e^{-\omega}},
#
# and
#
# .. math:: dx = \frac{(b-a)e^{-\omega}}{(1+e^{-\omega})^2}d\omega.
a = 0
b = 1
t = a + (b - a) / (1 + ex.exp(-omega))
dt = (b - a) * ex.exp(-omega) * (1 + ex.exp(-omega)) ** -2
integrand = t * t
z = ex.Integrate(integrand * dt / (b - a), 'omega')
z.getValue_c(prepareIds=True)

# %%
# Expressions using a database
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [10, 20, 30, 40, 50],
        'Variable2': [100, 200, 300, 400, 500],
        'Choice': [2, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
my_data = db.Database('test', df)

# %%
# Linear utility:  it defines a linear conbinations of parameters are variables.
beta1 = ex.Beta('beta1', 10, None, None, 0)
beta2 = ex.Beta('beta2', 20, None, None, 0)
v1 = ex.Variable('Variable1')
v2 = ex.Variable('Variable2')

# %%
listOfTerms = [
    (beta1, v1),
    (beta2, v2),
]
z = ex.bioLinearUtility(listOfTerms)
z.getValue_c(database=my_data, prepareIds=True)

# %%
# It is equivalent to the following, but implemented in a more efficient way.
z = beta1 * v1 + beta2 * v2
z.getValue_c(database=my_data, prepareIds=True)

# %%
# Monte Carlo: we approximate the integral
#
# .. math:: \int_0^1 x^2 dx=\frac{1}{3}
#
# using Monte-Carlo integration. As draws require a database, it is
# calculated for each entry in the database.
draws = ex.bioDraws('draws', 'UNIFORM')
z = ex.MonteCarlo(draws * draws)
z.getValue_c(database=my_data, prepareIds=True)

# %%
# Panel Trajectory: we first calculate a quantity for each entry in the database.
v1 = ex.Variable('Variable1')
v2 = ex.Variable('Variable2')
p = v1 / (v1 + v2)
p.getValue_c(database=my_data, prepareIds=True)

# %%
# We now declare the data as "panel", based on the identified
# `Person`. It means that the first three rows correspond to a
# sequence of three observations for individual 1, and the last two, a
# sequence of two observations for individual 2. The panel trajectory
# calculates the expression for each row associated with an
# individual, and calculate the product.
my_data.panel('Person')

# %%
# In this case, we expect the following for individual 1:
0.09090909**3

# %%
# And the following for individual 2:
0.09090909**2

# %%
# We verify that it is indeed the case:
z = ex.PanelLikelihoodTrajectory(p)
z.getValue_c(database=my_data, prepareIds=True)


# %%
# More complex expressions
# ------------------------
# We set the number of draws for Monte-Carlo integration. It should be
# a large number. For the sake of computational efficiency, as this
# notebook is designed to illustrate the various function, we use a
# low value.

# %%
NUMBER_OF_DRAWS = 100

# %%
# We first create a small database
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [10, 20, 30, 40, 50],
        'Variable2': [100, 200, 300, 400, 500],
        'Choice': [2, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
df

# %%
my_data = db.Database('test', df)

# %%
# The following type of expression is a literal called Variable that
# corresponds to an entry in the database.
Person = ex.Variable('Person')
Variable1 = ex.Variable('Variable1')
Variable2 = ex.Variable('Variable2')
Choice = ex.Variable('Choice')
Av1 = ex.Variable('Av1')
Av2 = ex.Variable('Av2')
Av3 = ex.Variable('Av3')

# %%
# It is possible to add a new column to the database, that creates a
# new variable that can be used in expressions.
newvar_b = my_data.DefineVariable('newvar_b', Variable1 + Variable2)
my_data.data

# %%
# It is equivalent to the following Pandas statement.
my_data.data['newvar_p'] = my_data.data['Variable1'] + my_data.data['Variable2']
my_data.data

# %%
# **Do not use chaining comparison expressions with Biogeme. Not only
# it does not provide the expected expression, but it does not
# trigger a warning or an exception.**
my_expression = 200 <= Variable2 <= 400
print(my_expression)

# %%
# The reason is that Python executes `200 <= Variable2 <= 400` as
# `(200 <= Variable2) and (Variable2 <= 400)`. The `and` operator
# cannot be overloaded in Python. Therefore, it does not return a
# Biogeme expression. Note that Pandas does not allow chaining either,
# and has implemented a `between` function instead.
my_data.data['chaining_p'] = my_data.data['Variable2'].between(200, 400)
my_data.data

# %%
# The following type of expression is another literal, corresponding
# to an unknown parameter. Note that the value is just a starting
# value for the algorithm.

beta1 = ex.Beta('beta1', 0.2, None, None, 0)
beta2 = ex.Beta('beta2', 0.4, None, None, 0)

# %%
# The last argument allows to fix the value of the parameter to the
# value.
beta3 = ex.Beta('beta3', 1, None, None, 1)
beta4 = ex.Beta('beta4', 0, None, None, 1)

# %%
# Arithmetic operators are overloaded to allow standard manipulations
# of expressions.
expr0 = beta3 + beta4
print(expr0)

# %%
# The evaluation of expressions can be done in two ways. For simple
# expressions, the fonction `getValue`, implemented in Python, returns
# the value of the expression.
expr0.getValue()

# %%
# It is possible to modify the values of the parameters.
newvalues = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
expr0.change_init_values(newvalues)
expr0.getValue()

# %%
# Consider another expression:
#
# .. math:: e_1 = 2 \beta_1 - \frac{\exp(-\beta_2)}{\beta_2 (\beta_3
#     \geq \beta_4) + \beta_1 (\beta_3 < \beta_4)},
#
# where :math:`(\beta_2 \geq \beta_1)` equals 1 if :math:`\beta_2 \geq
# \beta_1` and 0 otherwise.
expr1 = 2 * beta1 - ex.exp(-beta2) / (
    beta2 * (beta3 >= beta4) + beta1 * (beta3 < beta4)
)
print(expr1)


# %%
# The function `getValue_c` is implemented in C++, and works for any
# expression. When use outside a specific context, the IDs must be
# explicitly prepared.
expr1.getValue_c(prepareIds=True)

# %%
# It actually calls the function `getValueAndDerivates`, and returns
# its first output (without calculating the derivatives).
f, g, h, bhhh = expr1.getValueAndDerivatives(prepareIds=True)

# %%
f

# %%
# We create a pandas DataFrame just to have a nicer display of the results.
pd.DataFrame(g)

# %%
pd.DataFrame(h)

# %%
pd.DataFrame(bhhh)

# %%
# Note that the BHHH matrix is the outer product of the gradient with itself.
pd.DataFrame(np.outer(g, g))

# %%
# If the derivatives are not needed, their calculation can be
# skipped. Here, we calculate the gradient, but not the hessian.
expr1.getValueAndDerivatives(gradient=True, hessian=False, bhhh=False, prepareIds=True)

# %%
# It can also generate a function that takes the value of the
# parameters as argument, and provides a tuple with the value of the
# expression and its derivatives. By default, it returns the value of
# the function, its gradient and its hessian.
the_function = expr1.createFunction()

# %%
# We evaluate it at one point...
the_function([1, 2])

# %%
# ... and at another point.
the_function([10, -2])

# %%
# We can use it to check the derivatives.
tools.checkDerivatives(the_function, [1, 2], logg=True)

# %%
# And it is possible to also obtain the BHHH matrix.
the_function = expr1.createFunction(bhhh=True)
the_function([1, 2])

# %%
# It can take a database as input, and evaluate the expression and its
# derivatives for each entry in the database.  In the following
# example, as no variable of the database is involved in the
# expression, the output of the expression is the same for each entry.
results = expr1.getValueAndDerivatives(database=my_data, aggregation=False)
print(len(results))

# %%
f_array, g_array, h_array, bhhh_array = results

for f, g, h, bhhh in zip(f_array, g_array, h_array, bhhh_array):
    print('******')
    print(f'{f=}')
    print(f'{g=}')
    print(f'{h=}')
    print(f'{bhhh=}')

# %%
# If `aggregation` is set to `True`, the results are accumulated as a sum.
f, g, h, bhhh = expr1.getValueAndDerivatives(database=my_data, aggregation=True)
print(f'{f=}')
print(f'{g=}')
print(f'{h=}')
print(f'{bhhh=}')

# %%
# The following function scans the expression and extracts a dict with
# all free parameters.
expr1.set_of_elementary_expression(TypeOfElementaryExpression.FREE_BETA)

# %%
# Options can be set to extract free parameters, fixed parameters, or both.
expr1.set_of_elementary_expression(TypeOfElementaryExpression.FIXED_BETA)

# %%
expr1.set_of_elementary_expression(TypeOfElementaryExpression.BETA)

# %%
# It is possible also to extract an elementary expression from its name.
expr1.getElementaryExpression('beta2')

# %%
# Let's consider an expression involving two variables :math:`V_1` and
# :math:`V_2`:
#
# .. math:: e_2 = 2 \beta_1 V_1 - \frac{\exp(-\beta_2 V_2)}{\beta_2 (\beta_3
#     \geq \beta_4) + \beta_1 (\beta_3 < \beta_4)},
#
# where :math:`(\beta_2 \geq \beta_1)` equals 1 if :math:`\beta_2 \geq
# \beta_1` and 0 otherwise. Note that, in our example, the second term
# is numerically negligible with respect to the first one.

expr2 = 2 * beta1 * Variable1 - ex.exp(-beta2 * Variable2) / (
    beta2 * (beta3 >= beta4) + beta1 * (beta3 < beta4)
)
print(expr2)

# %%
# It is not a simple expression anymore, and only the function
# `getValue_c` can be invoked. If we try the `getValue` function, it
# raises an exception.
try:
    expr2.getValue()
except excep.BiogemeError as e:
    print(f'Exception raised: {e}')

# %%
# As the expression is called out of a specific context, it should be
# instructed to prepare its IDs. Note that if no database is
# provided, an exception is raised when the formula contains
# variables. Indeed, the values of these variables cannot be found
# anywhere.
try:
    expr2.getValue_c(prepareIds=True)
except excep.BiogemeError as e:
    print(f'Exception raised: {e}')

# %%
expr2.getValue_c(database=my_data, aggregation=False, prepareIds=True)

# %%
# The following function extracts the names of the parameters
# apprearing in the expression.
expr2.set_of_elementary_expression(TypeOfElementaryExpression.BETA)

# %%
# The list of parameters can also be obtained in the form of a dictionary.
expr2.dict_of_elementary_expression(TypeOfElementaryExpression.BETA)

# %%
# The list of variables can also be obtained in the form of a dictionary.
expr2.dict_of_elementary_expression(TypeOfElementaryExpression.VARIABLE)

# %%
# or a set...
expr2.set_of_elementary_expression(TypeOfElementaryExpression.VARIABLE)

# %%
# Expressions are defined recursively, using a tree
# representation. The following function describes the type of the
# upper most node of the tree.
expr2.getClassName()

# %%
# The signature is a formal representation of the expression,
# assigning identifiers to each node of the tree, and representing
# them starting from the leaves. It is easy to parse, and is passed to
# the C++ implementation.

# %%
# As the expression is used out of a specific context, it must be
# prepared before using it.
expr2.prepare(database=my_data, numberOfDraws=0)
expr2.getStatusIdManager()
print(expr2)

# %%
expr2.getSignature()

# %%
# The elementary expressions are
#
# - free parameters,
# - fixed parameters,
# - random variables (for numerical integration),
# - draws (for Monte-Carlo integration), and
# - variables from the database.
#
# The following function extracts all elementary expressions from a
# list of formulas, give them a unique numbering, and return them
# organized by group, as defined above (with the exception of the
# variables, that are directly available in the database).
collection_of_formulas = [expr1, expr2]
formulas = IdManager(collection_of_formulas, my_data, None)


# %%
# Unique numbering for all elementary expressions.
formulas.elementary_expressions.indices

# %%
formulas.free_betas

# %%
# Each elementary expression has two ids. One unique index across all
# elementary expressions, and one unique within each specific group.
[(i.elementaryIndex, i.betaId) for k, i in formulas.free_betas.expressions.items()]

# %%
formulas.free_betas.names

# %%
formulas.fixed_betas

# %%
[(i.elementaryIndex, i.betaId) for k, i in formulas.fixed_betas.expressions.items()]

# %%
formulas.fixed_betas.names

# %%
formulas.random_variables

# %%
# Monte Carlo integration is based on draws.
my_draws = ex.bioDraws('my_draws', 'UNIFORM')
expr3 = ex.MonteCarlo(my_draws * my_draws)
print(expr3)

# %%
# Note that draws are different from random variables, used for
# numerical integration.
expr3.set_of_elementary_expression(TypeOfElementaryExpression.RANDOM_VARIABLE)

# %%
# The following function reports the draws involved in an expression.
expr3.set_of_elementary_expression(TypeOfElementaryExpression.DRAWS)

# %%
# The following function checks if draws are defined outside
# MonteCarlo, and return their names.
wrong_expression = my_draws + ex.MonteCarlo(my_draws * my_draws)
wrong_expression.check_draws()

# %%
# Checking the correct expression returns an empty set.
expr3.check_draws()

# %%
# The expression is a Monte-Carlo integration.
expr3.getClassName()

# %%
# Note that the draws are associated with a database. Therefore, the
# evaluation of expressions involving Monte Carlo integration can only
# be done on a database. If none is provided, an exception is raised.
try:
    expr3.getValue_c(numberOfDraws=NUMBER_OF_DRAWS)
except excep.BiogemeError as e:
    print(f'Exception raised: {e}')

# %%
# Here is its value. It is an approximation of
#
# .. math:: \int_0^1 x^2 dx=\frac{1}{3}.
expr3.getValue_c(database=my_data, numberOfDraws=NUMBER_OF_DRAWS, prepareIds=True)

# %%
# Here is its signature.
expr3.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr3.getSignature()

# %%
# The same integral can be calculated using numerical integration,
# declaring a random variable.
omega = ex.RandomVariable('omega')

# %%
# Numerical integration calculates integrals between :math:`-\infty` and
# :math:`+\infty`. Here, the interval being :math:`[0,1]`, a change of variables
# is required.
a = 0
b = 1
x = a + (b - a) / (1 + ex.exp(-omega))
dx = (b - a) * ex.exp(-omega) * (1 + ex.exp(-omega)) ** (-2)
integrand = x * x
expr4 = ex.Integrate(integrand * dx / (b - a), 'omega')

# %%
# In this case, omega is a random variable.
expr4.dict_of_elementary_expression(TypeOfElementaryExpression.RANDOM_VARIABLE)

# %%
print(expr4)

# %%
# The folllowing function checks if random variables are defined
# outside an Integrate statement.
wrong_expression = x * x
wrong_expression.check_rv()

# %%
# The same function called from the correct expression returns an
# empty set.
expr4.check_rv()

# %%
# Calculating its value requires the C++ implementation.
expr4.getValue_c(my_data, prepareIds=True)

# %%
# We illustrate now the Elem function. It takes two arguments: a
# dictionary, and a formula for the key. For each entry in the
# database, the formula is evaluated, and its result identifies which
# formula in the dictionary should be evaluated.  Here is 'Person' is
# 1, the expression is
#
# .. math:: e_1=2 \beta_1 - \frac{\exp(-\beta_2)}{\beta_3 (\beta_2 \geq \beta_1)},
#
# and if 'Person' is 2, the expression is
#
# .. math:: e_2=2 \beta_1 V_1 - \frac{\exp(-\beta_2 V_2) }{ \beta_3 (\beta_2
#     \geq \beta_1)}.
#
# As it is a regular expression, it can be included in any
# formula. Here, we illustrate it by dividing the result by 10.
elemExpr = ex.Elem({1: expr1, 2: expr2}, Person)
expr5 = elemExpr / 10
print(expr5)

# %%
expr5.dict_of_elementary_expression(TypeOfElementaryExpression.VARIABLE)

# %%
# Note that `Variable1` and `Variable2` have previously been involved
# in another formula. Therefore, they have been numbered according to
# this formula, and this numbering is invalid for the new expression
# `expr5`. An error is triggered
try:
    expr5.getValue_c(database=my_data)
except excep.BiogemeError as e:
    print(e)

# %%
expr5.getValue_c(database=my_data, prepareIds=True)

# %%
testElem = ex.MonteCarlo(ex.Elem({1: my_draws * my_draws}, 1))

# %%
testElem.audit()

# %%
# The next expression is simply the sum of multiple expressions. The
# argument is a list of expressions.
expr6 = ex.bioMultSum([expr1, expr2, expr4])
print(expr6)

# %%
expr6.getValue_c(database=my_data, numberOfDraws=NUMBER_OF_DRAWS, prepareIds=True)

# %%
# We now illustrate how to calculate a logit model, that is
#
# .. math:: \frac{y_1 e^{V_1}}{y_0 e^{V_0}+y_1 e^{V_1}+y_2 e^{V_2}}
#
# where :math:`V_0=-\beta_1`, :math:`V_1=-\beta_2` and
# :math:`V_2=-\beta_1`, and :math:`y_i = 1`, :math:`i=1,2,3`.
V = {0: -beta1, 1: -beta2, 2: -beta1}
av = {0: 1, 1: 1, 2: 1}
expr7 = ex._bioLogLogit(V, av, 1)

# %%
expr7.getValue_c(prepareIds=True)

# %%
# If the alternative is not in the choice set, an exception is raised.
expr7_wrong = ex.LogLogit(V, av, 3)
try:
    expr7_wrong.getValue()
except excep.BiogemeError as e:
    print(f'Exception: {e}')

# %%
# It is actually better to use the C++ implementation, available in
# the module models.

# %%
expr8 = models.loglogit(V, av, 1)
expr8.getValue_c(database=my_data, prepareIds=True)

# %%
# As the result is a numpy array, it can be used for any
# calculation. Here, we show how to calculate the logsum.
for v in V.values():
    print(v.getValue_c(database=my_data, prepareIds=True))

# %%
logsum = np.log(
    np.sum(
        [np.exp(v.getValue_c(database=my_data, prepareIds=True)) for v in V.values()],
        axis=1,
    )
)
logsum

# %%
# It is possible to calculate the derivative of a formula with respect
# to a literal:
#
# .. math:: e_9=\frac{\partial e_8}{\partial \beta_2}.
expr9 = ex.Derive(expr8, 'beta2')
expr9.getValue_c(database=my_data, prepareIds=True)

# %%
expr9.elementaryName

# %%
# Biogeme also provides an approximation of the CDF of the normal
# distribution:
#
# .. math:: e_{10}= \frac{1}{{\sigma \sqrt {2\pi } }}\int_{-\infty}^t
#     e^{{{ - \left( {x - \mu } \right)^2 } \mathord{\left/ {\vphantom
#     {{ - \left( {x - \mu } \right)^2 } {2\sigma ^2 }}} \right. }
#     {2\sigma ^2 }}}dx.
expr10 = ex.bioNormalCdf(Variable1 / 10 - 1)
expr10.getValue_c(database=my_data, prepareIds=True)

# %%
# Min and max operators are also available. To avoid any ambiguity
# with the Python operator, they are called bioMin and bioMax.
expr11 = ex.bioMin(expr5, expr10)
expr11.getValue_c(database=my_data, prepareIds=True)

# %%
expr12 = ex.bioMax(expr5, expr10)
expr12.getValue_c(database=my_data, prepareIds=True)

# %%
# For the sake of efficiency, it is possible to specify explicitly a
# linear function, where each term is the product of a parameter and a
# variable.
terms = [
    (beta1, ex.Variable('Variable1')),
    (beta2, ex.Variable('Variable2')),
    (beta3, ex.Variable('newvar_b')),
]
expr13 = ex.bioLinearUtility(terms)
expr13.getValue_c(database=my_data, prepareIds=True)

# %%
# In terms of specification, it is equivalent to the expression
# below. But the calculation of the derivatives is more efficient, as
# the linear structure of the specification is exploited.
expr13bis = beta1 * Variable1 + beta2 * Variable2 + beta3 * newvar_b
expr13bis.getValue_c(database=my_data, prepareIds=True)

# %%
# A Pythonic way to write a linear utility function.
variables = ['v1', 'v2', 'v3', 'cost', 'time', 'headway']
coefficients = {f'{v}': ex.Beta(f'beta_{v}', 0, None, None, 0) for v in variables}
terms = [coefficients[v] * ex.Variable(v) for v in variables]
util = sum(terms)
print(util)

# %%
# If the data is organized a panel data, it means that several rows
# correspond to the same individual. The expression
# `PanelLikelihoodTrajectory` calculates the product of the expression
# evaluated for each row. If Monte Carlo integration is involved, the
# same draws are used for each them.

# %%
# Our database contains 5 observations.
my_data.getSampleSize()

# %%
my_data.panel('Person')

# %%
# Once the data has been labeled as "panel", it is considered that
# there are only two series of observations, corresponding to each
# person. Each of these observations is associated with several rows
# of observations.

# %%
my_data.getSampleSize()

# %%
# If we try to evaluate again the integral :math:`\int_0^1 x^2
# dx=\frac{1}{3}`, an exception is raised.
try:
    expr3.getValue_c(database=my_data)
except excep.BiogemeError as e:
    print(f'Exception: {e}')

# %%
# This is detected by the `audit` function, called before the
# expression is evaluated.
expr3.audit(database=my_data)

# %%
# We now evaluate an expression for panel data.
c1 = ex.bioDraws('draws1', 'NORMAL_HALTON2')
c2 = ex.bioDraws('draws2', 'NORMAL_HALTON2')
U1 = ex.Beta('beta1', 0, None, None, 0) * Variable1 + 10 * c1
U2 = ex.Beta('beta2', 0, None, None, 0) * Variable2 + 10 * c2
U3 = 0
U = {1: U1, 2: U2, 3: U3}
av = {1: Av1, 2: Av2, 3: Av3}
expr14 = ex.log(
    ex.MonteCarlo(ex.PanelLikelihoodTrajectory(models.logit(U, av, Choice)))
)

# %%
expr14.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr14

# %%
expr14.getValue_c(database=my_data, numberOfDraws=NUMBER_OF_DRAWS, prepareIds=True)

# %%
expr14.getValueAndDerivatives(
    database=my_data,
    numberOfDraws=NUMBER_OF_DRAWS,
    gradient=True,
    hessian=True,
    aggregation=False,
)

# %%
expr14.getValueAndDerivatives(
    database=my_data,
    numberOfDraws=NUMBER_OF_DRAWS,
    gradient=True,
    hessian=True,
    aggregation=True,
)

# %%
# A Python function can also be obtained for this expression. Note
# that it is available only for the aggregated version, summing over
# the database.

# %%
the_function = expr14.createFunction(
    database=my_data, numberOfDraws=NUMBER_OF_DRAWS, gradient=True, hessian=True
)

# %%
the_function([0, 0])

# %%
the_function([0.1, 0.1])

# %%
# It is possible to fix the value of some (or all) beta parameters
print(expr14)

# %%
expr14.fix_betas({'beta2': 0.123})

# %%
print(expr14)

# %%
# The name of the parameter can also be changed while fixing its value.

# %%
expr14.fix_betas({'beta2': 123}, prefix='prefix_', suffix='_suffix')

# %%
print(expr14)

# %%
# It can also be renamed using the following function.

# %%
expr14.rename_elementary(['prefix_beta2_suffix'], prefix='PREFIX_', suffix='_SUFFIX')

# %%
print(expr14)

# %%
# Signatures
# ----------

# %%
# The Python library communicates the expressions to the C++ library
# using a syntax called a "signature". We describe and illustrate now
# the signature for each expression. Each expression is identified by
# an identifier provided by Python using the function 'id'.

# %%
id(expr1)

# %%
# Numerical expression
# ++++++++++++++++++++

# %%
# <Numeric>{identifier},0.0
ex.Numeric(0).getSignature()

# %%
# Beta parameters
# +++++++++++++++

# %%
# <Beta>{identifier}"name"[status],uniqueId,betaId'
# where
#
# - status is 0 for free parameters, and non zero for fixed
#   parameters,
# - uniqueId is a unique index given by Biogeme to all elementary
#   expressions,
# - betaId is a unique index given by Biogeme to all free parameters,
#   and to all fixed parameters.

# %%
# As the signature requires an Id, we need to prepare the expression
# first.

# %%
beta1.prepare(database=my_data, numberOfDraws=0)
beta1.getSignature()

# %%
beta3.prepare(database=my_data, numberOfDraws=0)
beta3.getSignature()

# %%
# Variables
# +++++++++

# %%
# <Variable>{identifier}"name",uniqueId,variableId
# where
#
# - uniqueId is a unique index given by Biogeme to all elementary
#   expressions,
# - variableId is a unique index given by Biogeme to all variables.

# %%
Variable1.getSignature()

# %%
# Random variables
# ++++++++++++++++

# %%
# <RandomVariable>{identifier}"name",uniqueId,randomVariableId
# where
#
# - uniqueId is a unique index given by Biogeme to all elementary
#   expressions,
# - randomVariableId is a unique index given by Biogeme to all random
#   variables.

# %%
omega.prepare(database=my_data, numberOfDraws=0)
omega.getSignature()

# %%
# Draws
# +++++

# %%
# <bioDraws>{identifier}"name",uniqueId,drawId
# where
#
# - uniqueId is a unique index given by Biogeme to all elementary
#   expressions,
# - drawId is a unique index given by Biogeme to all draws.

# %%
my_draws.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
my_draws.getSignature()

# %%
# General expression
# ++++++++++++++++++
# `<operator>{identifier}(numberOfChildren),idFirstChild,idSecondChild,idThirdChild,`
# etc...
#
# where the number of identifiers given after the comma matches the
# reported number of children.
#
# Specific examples are reported below.

# %%
# Binary operator
# ///////////////

# %%
# `<code><operator>{identifier}(2),idFirstChild,idSecondChild </code>`
# where operator is one of:
#
#     - `Plus`
#     - `Minus`
#     - `Times`
#     - `Divide`
#     - `Power`
#     - `bioMin`
#     - `bioMax`
#     - `And`
#     - `Or`
#     - `Equal`
#     - `NotEqual`
#     - `LessOrEqual`
#     - `GreaterOrEqual`
#     - `Less`
#     - `Greater`

# %%
the_sum = beta1 + Variable1

# %%
the_sum.getSignature()

# %%
# Unary operator
# //////////////

# %%
# `<operator>{identifier}(1),idChild,`
# where operator is one of:
#
#     - `UnaryMinus`
#     - `MonteCarlo`
#     - `bioNormalCdf`
#     - `PanelLikelihoodTrajectory`
#     - `exp`
#     - `log`

# %%
m = -beta1

# %%
m.getSignature()

# %%
# LogLogit
# ////////

# %%
# <LogLogit>{identifier}(nbrOfAlternatives),chosenAlt,altNumber,utility,availability,altNumber,utility,availability, etc.

# %%
expr7.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr7.getSignature()

# %%
# Derive
# //////

# %%
# <Derive>{identifier},id of expression to derive,unique index of elementary expression

# %%
expr9.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr9.getSignature()

# %%
# Integrate
# /////////

# %%
# <Integrate>{identifier},id of expression to derive,index of random variable

# %%
expr4.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr4.getSignature()

# %%
# Elem
# ////

# %%
# <Elem>{identifier}(number_of_expressions),keyId,value1,expression1,value2,expression2, etc...
#
# where
#
# - keyId is the identifier of the expression calculating the key,
# - the number of pairs valuex,expressionx must correspond to the
#   value of number_of_expressions

# %%
elemExpr.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
elemExpr.getSignature()

# %%
# bioLinearUtility
# ////////////////

# %%
# <bioLinearUtility>{identifier}(numberOfTerms), beta1_exprId, beta1_uniqueId, beta1_name, variable1_exprId, variable1_uniqueId, variable1_name, etc...
#
# where 6 entries are provided for each term:
#
#     - beta1_exprId is the expression id of the beta parameter
#     - beta1_uniqueId is the unique id of the beta parameter
#     - beta1_name is the name of the parameter
#     - variable1_exprId is the expression id of the variable
#     - variable1_uniqueId is the unique id of the variable
#     - variable1_name is the name of the variable

# %%
expr13.prepare(database=my_data, numberOfDraws=NUMBER_OF_DRAWS)
expr13.getSignature()



"""Information about the version of Biogeme

:author: Michel Bierlaire
:date: Tue Mar 26 16:45:15 2019

"""

import datetime

from biogeme.deprecated import deprecated

# Release date
versionDate = f'{datetime.date.today()}'
AUTHOR = 'Michel Bierlaire'
URL_AUTHOR = 'http://people.epfl.ch/michel.bierlaire'
DEPARTMENT = 'Transport and Mobility Laboratory'
URL_DEPARTMENT = 'http://transp-or.epfl.ch'
UNIVERSITY = 'Ecole Polytechnique Fédérale de Lausanne (EPFL)'
URL_UNIVERSITY = 'http://www.epfl.ch'
URL_BIOGEME = 'http://biogeme.epfl.ch'
URL_FORUM = 'https://groups.google.com/d/forum/biogeme'

__version__ = '3.3.0a02'


def get_version() -> str:
    """
     Version of the software

    :return:  version number, and the release.
    :rtype: string
    """
    return __version__


@deprecated(get_version)
def getVersion() -> str:
    pass


def get_html() -> str:
    """Package information in HTML format

    :return: HTML code.
    :rtype: string

    """
    html = f'<p>biogeme {get_version()} [{versionDate}]</p>\n'
    html += (
        '<p><a href="https://www.python.org/" '
        'target="_blank">Python</a> package</p>\n'
    )
    html += (
        f'<p>Home page: <a href="{URL_BIOGEME}" '
        f'target="_blank">{URL_BIOGEME}</a></p>\n'
    )
    html += (
        f'<p>Submit questions to <a href="{URL_FORUM}" '
        f'target="_blank">{URL_FORUM}</a></p>\n'
    )
    html += f'<p><a href="{URL_AUTHOR}">'
    html += AUTHOR
    html += '</a>, <a href="'
    html += URL_DEPARTMENT
    html += '">'
    html += DEPARTMENT
    html += '</a>, <a href="'
    html += URL_UNIVERSITY
    html += '">'
    html += UNIVERSITY.encode('ascii', 'xmlcharrefreplace').decode()
    html += '</a></p>\n'
    return html


@deprecated(get_html)
def getHtml() -> str:
    pass


def get_text() -> str:
    """Package information in text format

    :return: package information
    :rtype: string
    """

    text = f'biogeme {get_version()} [{versionDate}]\n'
    text += f'Home page: {URL_BIOGEME}\n'
    text += f'Submit questions to {URL_FORUM}\n'
    text += f'{AUTHOR}, {DEPARTMENT}, {UNIVERSITY}\n'
    return text


@deprecated(get_text)
def getText() -> str:
    pass


def get_latex() -> str:
    """Package information in LaTeX format

    :return: LaTeX comments
    :rtype: string
    """
    latex = f'%% biogeme {get_version()} [{versionDate}]\n'
    latex += f'%% Home page: {URL_BIOGEME}\n'
    latex += f'%% Submit questions to {URL_FORUM}\n'
    latex += f'%% {AUTHOR}, {DEPARTMENT}, {UNIVERSITY}\n'
    return latex


@deprecated(get_latex)
def getLaTeX() -> str:
    pass



from .database import countNumberOfGroups, count_number_of_groups, flatten_database
from .derivatives import (
    CheckDerivativesResults,
    check_derivatives,
    findiff_g,
    findiff_h,
)
from .files import TemporaryFile, create_backup, is_valid_filename
from .formatting import format_memory_size
from .likelihood_ratio import LRTuple, likelihood_ratio_test
from .primes import calculate_prime_numbers, get_prime_numbers
from .serialize_numpy import safe_deserialize_array, safe_serialize_array
from .simulate import simulate
from .time import format_timedelta
from .unique_ids import ModelNames, generate_unique_ids, unique_product



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta
from biogeme.parameters import Parameters

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


# Associate the availability conditions with the alternatives

av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


class test_01(unittest.TestCase):
    def setUp(self) -> None:
        """Create the configuration .py"""
        self.scipy_configuration = Parameters()
        self.scipy_configuration.set_value(
            name='optimization_algorithm', value='scipy', section='Estimation'
        )
        self.scipy_configuration.set_value(name='seed', value=10, section='MonteCarlo')

        self.ls_configuration = Parameters()
        self.ls_configuration.set_value(
            name='optimization_algorithm', value='LS-newton', section='Estimation'
        )
        self.ls_configuration.set_value(name='seed', value=10, section='MonteCarlo')

        self.tr_configuration = Parameters()
        self.tr_configuration.set_value(
            name='optimization_algorithm', value='TR-newton', section='Estimation'
        )
        self.tr_configuration.set_value(name='seed', value=10, section='MonteCarlo')

        self.simple_bounds_configuration = Parameters()
        self.simple_bounds_configuration.set_value(
            name='optimization_algorithm', value='simple_bounds', section='Estimation'
        )
        self.simple_bounds_configuration.set_value(
            name='seed', value=10, section='MonteCarlo'
        )

    def testEstimationScipy(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=self.scipy_configuration,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
            bootstrap_samples=10,
        )
        biogeme.model_name = 'test_01'
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.final_log_likelihood, -5331.252, 2)

    def _testEstimationLineSearch(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=self.ls_configuration,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
            bootstrap_samples=10,
        )
        biogeme.model_name = 'test_01'
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.final_log_likelihood, -5331.252, 2)

    def testEstimationTrustRegion(self):
        logprob = models.loglogit(V, av, CHOICE)
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=self.tr_configuration,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
            bootstrap_samples=10,
        )
        biogeme.model_name = 'test_01'
        results = biogeme.estimate(run_bootstrap=True)
        self.assertAlmostEqual(results.final_log_likelihood, -5331.252, 2)


if __name__ == '__main__':
    unittest.main()



import glob
import logging
import os
import shutil
import tempfile
import uuid
from os import path

logger = logging.getLogger(__name__)


class TemporaryFile:
    def __enter__(self):
        self.dir = tempfile.mkdtemp()
        self.name = str(uuid.uuid4())
        self.fullpath = path.join(self.dir, self.name)
        return self  # <--- return the object itself

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.dir)

    def __str__(self):
        return self.fullpath

    def remove(self):
        shutil.rmtree(self.dir)


def is_valid_filename(filename: str) -> tuple[bool, str]:
    """Verifies is a string is a valid filename"""
    # Check for zero length or None
    if not filename or len(filename) == 0:
        return False, 'Name is empty'

    # Define invalid characters for file names (common for Windows, Linux, and macOS)
    invalid_chars = '<>:"/\\|?*'
    if any(char in invalid_chars for char in filename):
        return False, f'Name contains one invalid char: {invalid_chars}'

    # Check for system reserved names in Windows
    if os.name == 'nt':  # nt means Windows
        reserved_names = {
            "CON",
            "PRN",
            "AUX",
            "NUL",
            "COM1",
            "COM2",
            "COM3",
            "COM4",
            "COM5",
            "COM6",
            "COM7",
            "COM8",
            "COM9",
            "LPT1",
            "LPT2",
            "LPT3",
            "LPT4",
            "LPT5",
            "LPT6",
            "LPT7",
            "LPT8",
            "LPT9",
        }
        if filename.upper() in reserved_names:
            return False, f'Name is a reserved name: {reserved_names}'

    # Check filename length (common limit is 255 for many file systems, but this is OS dependent)
    if len(filename) > 255:
        return False, f'The length of the filename exceeds 255: {len(filename)}'

    return True, ''


def create_backup(filename: str, rename: bool = True) -> str:
    """If the file exist, we create a backup copy.

    :param filename: name of the file
    :param rename: if True, the file is renamed. If False, a copy is made.
    """
    original_base_name, original_extension = os.path.splitext(filename)

    # Check if the file already exists
    if os.path.exists(filename):
        # Initialize a counter for potential file name conflicts
        counter = 1
        # Generate a new file name by appending a number
        # Keep incrementing the number if the file already exists
        while True:
            new_name = f'{original_base_name}_{counter}{original_extension}'
            if not os.path.exists(new_name):
                break  # Exit loop if the file name does not exist
            counter += 1  # Increment the counter and try again

        # Rename the old file to the new file name
        if rename:
            os.rename(filename, new_name)
            logger.info(f'File {filename} has been renamed {new_name}')
        else:
            shutil.copy(filename, new_name)
            logger.info(f'File {filename} has been copied as {new_name}')
        return new_name
    logger.info(f'File {filename} does not exist. No backup has been generated')


def files_of_type(extension: str, name: str, all_files: bool = False) -> list[str]:
    """Identify the list of files with a given extension in the
    local directory

    :param extension: extension of the requested files (without
        the dot): 'yaml', or 'html'
    :param name: filename, without extension
    :param all_files: if all_files is False, only files containing
        the name of the model are identified. If all_files is
        True, all files with the requested extension are
        identified.

    :return: list of files with the requested extension.
    """
    if all_files:
        pattern = f"*.{extension}"
        return glob.glob(pattern)
    pattern1 = f"{name}.{extension}"
    pattern2 = f"{name}~*.{extension}"
    files = glob.glob(pattern1) + glob.glob(pattern2)
    return files



"""Utilities to perform various checks

Michel Bierlaire
Fri Apr 5 11:09:39 2024
"""

from typing import Any


def check_consistency_of_named_dicts(
    named_dicts: dict[str, dict[Any, Any]]
) -> tuple[bool, str | None]:
    """Verify that all dictionaries have the same set of keys. If not, report the inconsistencies.

    :param named_dicts: A dictionary where each key is a name (str) and each value is a dictionary to check.
    :return: True if keys are consistent across all dictionaries, False otherwise. If not consistent,
             returns a report detailing the inconsistencies.
    """
    # Initialize variables
    dict_names = list(named_dicts.keys())
    keys = [set(d.keys()) for d in named_dicts.values()]
    consistent = True
    inconsistencies = []

    # Reference keys for comparison
    reference_keys = keys[0]
    reference_name = dict_names[0]

    # Compare each set of keys with the first to check for consistency
    for i, (name, key_set) in enumerate(zip(dict_names[1:], keys[1:]), start=1):
        if key_set != reference_keys:
            consistent = False
            missing_in_ref = reference_keys - key_set
            extra_in_ref = key_set - reference_keys
            if missing_in_ref:
                inconsistencies.append(
                    f'{name} missing keys compared to {reference_name}: {missing_in_ref}.'
                )
            if extra_in_ref:
                inconsistencies.append(
                    f'{name} has extra keys compared to {reference_name}: {extra_in_ref}.'
                )

    # Compile the report if there are inconsistencies
    report = None
    if not consistent:
        report = "Inconsistencies found among dictionaries:\n" + "\n".join(
            inconsistencies
        )

    return consistent, report


def validate_dict_types(
    parameter: dict, name: str, value_type: type, key_type: type = int
) -> None:
    """Validate the types of keys and values in a dictionary.

    :param parameter: The dictionary parameter to validate.
    :param name: The name of the parameter for error messages.
    :param value_type: The expected type of the dictionary values.
    :param key_type: The expected type of the dictionary keys. Default is int.
    :raises TypeError: If key or value types do not match expectations.
    """
    if not isinstance(parameter, dict):
        raise TypeError(
            f"{name} must be a dict, got {type(parameter).__name__} instead."
        )
    for key, value in parameter.items():
        if not isinstance(key, key_type):
            raise TypeError(
                f"Keys in {name} must be of type {key_type.__name__}, found type {type(key).__name__}."
            )
        if not (isinstance(value, value_type) or value is None):
            raise TypeError(
                f"Values in {name} must be of type {value_type.__name__} or None, found type {type(value).__name__}."
            )
